import os
import sys


# In Python versions prior to 3.4, __file__ returns a relative path. This path
# is fixed at load time, so if the program later cd's (as we do in tests, at
# least) __file__ is no longer valid. As a workaround, compute the absolute
# path at load time.
MODULE_ROOT = os.path.abspath(os.path.dirname(__file__))


def makedirs(path):
    '''os.makedirs() has an exist_ok param, but it still throws errors when the
    path exists with non-default permissions. This isn't fixed until 3.4.
    Pathlib won't be getting an exist_ok param until 3.5.'''
    path = str(path)  # compatibility with pathlib
    if not os.path.exists(path):
        os.makedirs(path)


def is_fancy_terminal():
    '''The Windows terminal does not support most of the fancy things we want
    to do with colors and formatting. This is a quick and dirty way to make
    sure we default to simple output on Windows.'''
    return sys.stdout.isatty() and os.name != 'nt'



import asyncio
import textwrap

from .cache import compute_key, MergeConflictError


@asyncio.coroutine
def merge_imports_tree(cache, imports, target_trees, base_tree=None):
    '''Take an Imports struct and a dictionary of resolved trees and merge the
    unified imports tree. If base_tree is supplied, merge that too. There are a
    couple reasons for structuring this function the way it is:
        - We want to cache merged trees, so that we don't have to do expensive
          git operations just to check whether a module is in cache.
        - We want tree merging to know about target names, so that it can write
          good error messages when there are conflicts.
        - We need to use this for both toplevel imports and recursive module
          imports.
    '''
    key = _cache_key(imports, target_trees, base_tree)
    if key in cache.keyval:
        return cache.keyval[key]
    # We always want to merge imports in the same order, so that any conflicts
    # we run into will be deterministic. Sort the imports alphabetically by
    # target name.
    unified_tree = base_tree or (yield from cache.get_empty_tree())
    for target, paths in imports.items():
        for path in paths:
            try:
                unified_tree = yield from cache.merge_trees(
                    unified_tree, target_trees[target], path)
            except MergeConflictError as e:
                message = 'Merge conflict in import "{}" at "{}":\n\n{}'
                e.message = message.format(
                    target, path, textwrap.indent(e.message, '  '))
                raise
    cache.keyval[key] = unified_tree
    return unified_tree


def _cache_key(imports, target_trees, base_tree):
    tree_paths = tuple((target_trees[target], paths) for
                       target, paths in imports.items())
    return compute_key({
        'base_tree': base_tree,
        'tree_paths': tree_paths,
    })



from pathlib import PurePosixPath
import re

from .error import PrintableError


UNESCAPED_STAR_EXPR = (
    r'(?<!\\)'    # negative lookbehind assertion for more backslashes
    r'(?:\\\\)*'  # non-capturing group of an even number of backslashes
    r'\*'         # literal *
)


def contains_unescaped_stars(glob):
    return re.search(UNESCAPED_STAR_EXPR, glob) is not None


def unglobbed_prefix(glob):
    '''Returns all the path components, starting from the beginning, up to the
    first one with any kind of glob. So for example, if glob is 'a/b/c*/d',
    return 'a/b'.'''
    parts = []
    for part in PurePosixPath(glob).parts:
        if contains_unescaped_stars(part):
            break
        else:
            parts.append(part)
    return str(PurePosixPath(*parts)) if parts else ''


def _split_on_indices(s, indices):
    start = 0
    for i in indices:
        yield s[start:i]
        start = i + 1
    yield s[start:]


def split_on_stars_interpreting_backslashes(s):
    r'''We don't want to do in-place substitutions of a regex for *, because we
    need to be able to regex-escape the rest of the string. Instead, we split
    the string on *'s, so that the rest can be regex-escaped and then rejoined
    with the right regex. While we're doing this, check for backslash-escaped
    *'s and \'s, and leave them in as literals (to be regex-escaped in the next
    step).'''

    star_indices = [match.end() - 1 for match in
                    re.finditer(UNESCAPED_STAR_EXPR, s)]
    literalized_parts = [part.replace(r'\*', '*').replace(r'\\', '\\')
                         for part in _split_on_indices(s, star_indices)]
    return literalized_parts


def glob_to_path_regex(glob):
    '''Supports * and **. Backslashes can escape stars or other backslashes. As
    in pathlib, ** may not adjoin any characters other than slash. Unlike
    pathlib, because we're not talking to the actual filesystem, ** will match
    files as well as directories. Paths get canonicalized before they're
    converted, so duplicate and trailing slashes get dropped. You should make
    sure the other paths you try to match are in canonical (Posix) form as
    well.'''

    canonical_glob = str(PurePosixPath(glob))

    # The final regex starts with ^ and ends with $ to force it to match the
    # whole path.
    regex = '^'
    components = canonical_glob.split('/')
    for i, component in enumerate(components):
        if component == '**':
            if i == len(components) - 1:
                raise GlobError(
                    glob, '** may not be the last component in a path.')
            else:
                regex += r'(?:[^/]+/)*'
        elif '**' in component:
            raise GlobError(glob, '** must be an entire path component.')
        else:
            if component == '*':
                # A lone * may not match empty.
                regex += r'[^/]+'
            else:
                # A * with other characters may match empty. Escape all other
                # regex special characters.
                star_parts = split_on_stars_interpreting_backslashes(component)
                escaped_parts = map(re.escape, star_parts)
                regex += r'[^/]*'.join(escaped_parts)
            # Add a trailing slash for every component except **.
            if i < len(components) - 1:
                regex += '/'

    regex += '$'
    return regex


class GlobError(PrintableError):
    def __init__(self, glob, message):
        new_message = 'Glob error in "{}": {}'.format(glob, message)
        super().__init__(new_message)



import asyncio
import collections
import os
from pathlib import Path
import tempfile

from . import cache
from . import compat
from .error import PrintableError
from . import display
from .keyval import KeyVal
from . import parser
from . import plugin


@asyncio.coroutine
def Runtime(args, env):
    'This is the async constructor for the _Runtime class.'
    r = _Runtime(args, env)
    yield from r._init_cache()
    return r


class _Runtime:
    def __init__(self, args, env):
        "Don't instantiate this class directly. Use the Runtime() constructor."
        self._set_paths(args, env)

        compat.makedirs(self.state_dir)

        self._tmp_root = os.path.join(self.state_dir, 'tmp')
        compat.makedirs(self._tmp_root)

        self.overrides = KeyVal(os.path.join(self.state_dir, 'overrides'),
                                self._tmp_root)
        self._used_overrides = set()

        self.force = args.get('--force', False)
        if args['--quiet'] and args['--verbose']:
            raise PrintableError(
                "Peru can't be quiet and verbose at the same time.")
        self.quiet = args['--quiet']
        self.verbose = args['--verbose']
        self.no_overrides = args.get('--no-overrides', False)
        self.no_cache = args.get('--no-cache', False)

        # Use a semaphore (a lock that allows N holders at once) to limit the
        # number of fetches that can run in parallel.
        num_fetches = _get_parallel_fetch_limit(args)
        self.fetch_semaphore = asyncio.BoundedSemaphore(num_fetches)

        # Use locks to make sure the same cache keys don't get double fetched.
        self.cache_key_locks = collections.defaultdict(asyncio.Lock)

        # Use a different set of locks to make sure that plugin cache dirs are
        # only used by one job at a time.
        self.plugin_cache_locks = collections.defaultdict(asyncio.Lock)

        self.display = get_display(args)

    @asyncio.coroutine
    def _init_cache(self):
        self.cache = yield from cache.Cache(self.cache_dir)

    def _set_paths(self, args, env):
        explicit_peru_file = args['--file']
        explicit_sync_dir = args['--sync-dir']
        explicit_basename = args['--file-basename']
        if explicit_peru_file and explicit_basename:
            raise CommandLineError(
                'Cannot use both --file and --file-basename at the same time.')
        if explicit_peru_file and explicit_sync_dir:
            self.peru_file = explicit_peru_file
            self.sync_dir = explicit_sync_dir
        elif explicit_peru_file or explicit_sync_dir:
            raise CommandLineError(
                'If the --file or --sync-dir is set, '
                'the other must also be set.')
        else:
            basename = explicit_basename or parser.DEFAULT_PERU_FILE_NAME
            self.peru_file = find_project_file(os.getcwd(), basename)
            self.sync_dir = os.path.dirname(self.peru_file)
        self.state_dir = (args['--state-dir'] or
                          os.path.join(self.sync_dir, '.peru'))
        self.cache_dir = (args['--cache-dir'] or env.get('PERU_CACHE_DIR') or
                          os.path.join(self.state_dir, 'cache'))

    def tmp_dir(self):
        dir = tempfile.TemporaryDirectory(dir=self._tmp_root)
        return dir

    def get_plugin_context(self):
        return plugin.PluginContext(
            # Plugin cwd is always the directory containing peru.yaml, even if
            # the sync_dir has been explicitly set elsewhere. That's because
            # relative paths in peru.yaml should respect the location of that
            # file.
            cwd=str(Path(self.peru_file).parent),
            plugin_cache_root=self.cache.plugins_root,
            parallelism_semaphore=self.fetch_semaphore,
            plugin_cache_locks=self.plugin_cache_locks,
            tmp_root=self._tmp_root)

    def set_override(self, name, path):
        if not os.path.isabs(path):
            # We can't store relative paths as given, because peru could be
            # running from a different working dir next time. But we don't want
            # to absolutify everything, because the user might want the paths
            # to be relative (for example, so a whole workspace can be moved as
            # a group while preserving all the overrides). So reinterpret all
            # relative paths from the project root.
            path = os.path.relpath(path, start=self.sync_dir)
        self.overrides[name] = path

    def get_override(self, name):
        if self.no_overrides or name not in self.overrides:
            return None
        path = self.overrides[name]
        if not os.path.isabs(path):
            # Relative paths are stored relative to the project root.
            # Reinterpret them relative to the cwd. See the above comment in
            # set_override.
            path = os.path.relpath(os.path.join(self.sync_dir, path))
        return path

    def mark_override_used(self, name):
        '''Marking overrides as used lets us print a warning when an override
        is unused.'''
        self._used_overrides.add(name)

    def print_overrides(self):
        if self.quiet or self.no_overrides:
            return
        names = sorted(self.overrides)
        if not names:
            return
        self.display.print('syncing with overrides:')
        for name in names:
            self.display.print('  {}: {}'.format(
                name, self.get_override(name)))

    def warn_unused_overrides(self):
        if self.quiet or self.no_overrides:
            return
        unused_names = set(self.overrides) - self._used_overrides
        if not unused_names:
            return
        self.display.print('WARNING unused overrides:')
        for name in sorted(unused_names):
            self.display.print('  ' + name)


def find_project_file(start_dir, basename):
    '''Walk up the directory tree until we find a file of the given name.'''
    prefix = os.path.abspath(start_dir)
    while True:
        candidate = os.path.join(prefix, basename)
        if os.path.isfile(candidate):
            return candidate
        if os.path.exists(candidate):
            raise PrintableError(
                "Found {}, but it's not a file.".format(candidate))
        if os.path.dirname(prefix) == prefix:
            # We've walked all the way to the top. Bail.
            raise PrintableError("Can't find " + basename)
        # Not found at this level. We must go...shallower.
        prefix = os.path.dirname(prefix)


def _get_parallel_fetch_limit(args):
    jobs = args.get('--jobs')
    if jobs is None:
        return plugin.DEFAULT_PARALLEL_FETCH_LIMIT
    try:
        parallel = int(jobs)
        if parallel <= 0:
            raise PrintableError('Argument to --jobs must be 1 or more.')
        return parallel
    except:
        raise PrintableError('Argument to --jobs must be a number.')


def get_display(args):
    if args['--quiet']:
        return display.QuietDisplay()
    elif args['--verbose']:
        return display.VerboseDisplay()
    elif compat.is_fancy_terminal():
        return display.FancyDisplay()
    else:
        return display.QuietDisplay()


class CommandLineError(PrintableError):
    pass



import asyncio
import os
from pathlib import Path

from .async import gather_coalescing_exceptions
from . import compat
from .error import error_context
from .merge import merge_imports_tree


@asyncio.coroutine
def checkout(runtime, scope, imports, path):
    imports_tree = yield from get_imports_tree(runtime, scope, imports)
    last_imports_tree = _get_last_imports(runtime)
    index = _last_imports_index(runtime)
    yield from runtime.cache.export_tree(
        imports_tree, path, last_imports_tree, force=runtime.force,
        previous_index_file=index)
    _set_last_imports(runtime, imports_tree)


@asyncio.coroutine
def get_imports_tree(runtime, scope, imports, base_tree=None):
    target_trees = yield from get_trees(runtime, scope, imports.keys())
    imports_tree = yield from merge_imports_tree(
        runtime.cache, imports, target_trees, base_tree)
    return imports_tree


@asyncio.coroutine
def get_trees(runtime, scope, targets):
    futures = [get_tree(runtime, scope, target) for target in targets]
    trees = yield from gather_coalescing_exceptions(
        futures,
        runtime.display,
        verbose=runtime.verbose)
    return dict(zip(targets, trees))


@asyncio.coroutine
def get_tree(runtime, scope, target_str):
    module, rules = yield from scope.parse_target(runtime, target_str)
    context = 'target "{}"'.format(target_str)
    with error_context(context):
        tree = yield from module.get_tree(runtime)
        if module.default_rule:
            tree = yield from module.default_rule.get_tree(runtime, tree)
        for rule in rules:
            tree = yield from rule.get_tree(runtime, tree)
    return tree


def _last_imports_path(runtime):
    return Path(runtime.state_dir) / 'lastimports'


def _get_last_imports(runtime):
    last_imports_tree = None
    if _last_imports_path(runtime).exists():
        with _last_imports_path(runtime).open() as f:
            last_imports_tree = f.read()
    return last_imports_tree


def _set_last_imports(runtime, tree):
    if tree == _get_last_imports(runtime):
        # Don't modify the lastimports file if the imports haven't changed.
        # This lets you use it as a build stamp for Make.
        return
    compat.makedirs(_last_imports_path(runtime).parent)
    with _last_imports_path(runtime).open('w') as f:
        f.write(tree)


def _last_imports_index(runtime):
    return os.path.join(runtime.state_dir, 'lastimports.index')



#! /usr/bin/env python3

import os
import subprocess
import sys


def svn(*args, svn_dir=None, capture_output=False):
    # Avoid forgetting this arg.
    assert svn_dir is None or os.path.isdir(svn_dir)

    command = ['svn', '--non-interactive']
    command.extend(args)

    stdout = subprocess.PIPE if capture_output else None
    # Always let stderr print to the caller.
    process = subprocess.Popen(
        command, stdin=subprocess.DEVNULL, stdout=stdout, cwd=svn_dir,
        universal_newlines=True)
    output, _ = process.communicate()
    if process.returncode != 0:
        sys.exit(1)

    return output


def remote_head_rev(url):
    print('svn info', url)
    info = svn('info', url, capture_output=True).split('\n')
    for item in info:
        if item.startswith('Revision: '):
            return item.split()[1]

    print('svn revision info not found', file=sys.stderr)
    sys.exit(1)


def plugin_sync():
    # Just fetch the target revision and strip the metadata.
    # Plugin-level caching for Subversion is futile.
    svn(
        'export',
        '--force',
        '--revision',
        os.environ['PERU_MODULE_REV'] or 'HEAD',
        os.environ['PERU_MODULE_URL'],
        os.environ['PERU_SYNC_DEST'])


def plugin_reup():
    url = os.environ['PERU_MODULE_URL']
    rev = remote_head_rev(url)
    output_file = os.environ['PERU_REUP_OUTPUT']
    with open(output_file, 'w') as f:
        # Quote Subversion revisions to prevent integer intepretation.
        print('rev:', '"{}"'.format(rev), file=f)


command = os.environ['PERU_PLUGIN_COMMAND']
if command == 'sync':
    plugin_sync()
elif command == 'reup':
    plugin_reup()
else:
    raise RuntimeError('Unknown command: ' + repr(command))



import asyncio
from pathlib import PurePosixPath
import re

from . import cache
from .error import PrintableError
from . import glob


class Rule:
    def __init__(self, name, copy, move, executable, drop, pick, export):
        self.name = name
        self.copy = copy
        self.move = move
        self.executable = executable
        self.drop = drop
        self.pick = pick
        self.export = export

    def _cache_key(self, input_tree):
        return cache.compute_key({
            'input_tree': input_tree,
            'copy': self.copy,
            'move': self.move,
            'executable': self.executable,
            'drop': self.drop,
            'pick': self.pick,
            'export': self.export,
        })

    @asyncio.coroutine
    def get_tree(self, runtime, input_tree):
        key = self._cache_key(input_tree)

        # As with Module, take a lock on the cache key to avoid running the
        # same rule (or identical rules) twice with the same input.
        cache_lock = runtime.cache_key_locks[key]
        with (yield from cache_lock):
            if key in runtime.cache.keyval:
                return runtime.cache.keyval[key]

            tree = input_tree
            if self.copy:
                tree = yield from copy_files(runtime.cache, tree, self.copy)
            if self.move:
                tree = yield from move_files(runtime.cache, tree, self.move)
            if self.drop:
                tree = yield from drop_files(
                    runtime.cache, tree, self.drop)
            if self.pick:
                tree = yield from pick_files(runtime.cache, tree, self.pick)
            if self.executable:
                tree = yield from make_files_executable(
                    runtime.cache, tree, self.executable)
            if self.export:
                tree = yield from get_export_tree(
                    runtime.cache, tree, self.export)

            runtime.cache.keyval[key] = tree

        return tree


@asyncio.coroutine
def _copy_files_modifications(_cache, tree, paths_multimap):
    modifications = {}
    for source in paths_multimap:
        source_info_dict = yield from _cache.ls_tree(tree, source)
        if not source_info_dict:
            raise NoMatchingFilesError(
                'Path "{}" does not exist.'.format(source))
        source_info = list(source_info_dict.items())[0][1]
        for dest in paths_multimap[source]:
            # If dest is a directory, put the source inside dest instead of
            # overwriting dest entirely.
            dest_is_dir = False
            dest_info_dict = yield from _cache.ls_tree(tree, dest)
            if dest_info_dict:
                dest_info = list(dest_info_dict.items())[0][1]
                dest_is_dir = (dest_info.type == cache.TREE_TYPE)
            adjusted_dest = dest
            if dest_is_dir:
                adjusted_dest = str(PurePosixPath(dest) /
                                    PurePosixPath(source).name)
            modifications[adjusted_dest] = source_info
    return modifications


@asyncio.coroutine
def copy_files(_cache, tree, paths_multimap):
    modifications = yield from _copy_files_modifications(
        _cache, tree, paths_multimap)
    tree = yield from _cache.modify_tree(tree, modifications)
    return tree


@asyncio.coroutine
def move_files(_cache, tree, paths_multimap):
    # First obtain the copies from the original tree. Moves are not ordered but
    # happen all at once, so if you move a->b and b->c, the contents of c will
    # always end up being b rather than a.
    modifications = yield from _copy_files_modifications(
        _cache, tree, paths_multimap)
    # Now add in deletions, but be careful not to delete a file that just got
    # moved. Note that if "a" gets moved into "dir", it will end up at "dir/a",
    # even if "dir" is deleted (because modify_tree always modifies parents
    # before decending into children, and deleting a dir is a modification of
    # that dir's parent).
    for source in paths_multimap:
        if source not in modifications:
            modifications[source] = None
    tree = yield from _cache.modify_tree(tree, modifications)
    return tree


@asyncio.coroutine
def _get_glob_entries(_cache, tree, globs_list):
    matches = {}
    for glob_str in globs_list:
        # Do an in-memory match of all the paths in the tree against the
        # glob expression. As an optimization, if the glob is something
        # like 'a/b/**/foo', only list the paths under 'a/b'.
        regex = glob.glob_to_path_regex(glob_str)
        prefix = glob.unglobbed_prefix(glob_str)
        entries = yield from _cache.ls_tree(tree, prefix, recursive=True)
        found = False
        for path, entry in entries.items():
            if re.match(regex, path):
                matches[path] = entry
                found = True
        if not found:
            raise NoMatchingFilesError(
                '"{}" didn\'t match any files.'.format(glob_str))
    return matches


@asyncio.coroutine
def pick_files(_cache, tree, globs_list):
    picks = yield from _get_glob_entries(_cache, tree, globs_list)
    tree = yield from _cache.modify_tree(None, picks)
    return tree


@asyncio.coroutine
def drop_files(_cache, tree, globs_list):
    drops = yield from _get_glob_entries(_cache, tree, globs_list)
    for path in drops:
        drops[path] = None
    tree = yield from _cache.modify_tree(tree, drops)
    return tree


@asyncio.coroutine
def make_files_executable(_cache, tree, globs_list):
    entries = yield from _get_glob_entries(_cache, tree, globs_list)
    exes = {}
    for path, entry in entries.items():
        # Ignore directories.
        if entry.type == cache.BLOB_TYPE:
            exes[path] = entry._replace(mode=cache.EXECUTABLE_FILE_MODE)
    tree = yield from _cache.modify_tree(tree, exes)
    return tree


@asyncio.coroutine
def get_export_tree(_cache, tree, export_path):
    entries = yield from _cache.ls_tree(tree, export_path)
    if not entries:
        raise NoMatchingFilesError('Export path "{}" doesn\'t exist.'
                                   .format(export_path))
    entry = list(entries.values())[0]
    if entry.type != cache.TREE_TYPE:
        raise NoMatchingFilesError('Export path "{}" is not a directory.'
                                   .format(export_path))
    return entry.hash


class NoMatchingFilesError(PrintableError):
    pass



import asyncio

from .error import PrintableError


SCOPE_SEPARATOR = '.'
RULE_SEPARATOR = '|'


class Scope:
    '''A Scope holds the elements that are parsed out of a single peru.yaml
    file. This is kept separate from a Runtime, because recursive modules need
    to work with a Scope that makes sense to them, rather than a single global
    scope.'''

    def __init__(self, modules, rules):
        self.modules = modules
        self.rules = rules

    @asyncio.coroutine
    def parse_target(self, runtime, target_str):
        '''A target is a pipeline of a module into zero or more rules, and each
        module and rule can itself be scoped with zero or more module names.'''
        pipeline_parts = target_str.split(RULE_SEPARATOR)
        module = yield from self.resolve_module(
            runtime, pipeline_parts[0], target_str)
        rules = tuple((yield from self.resolve_rule(runtime, part))
                      for part in pipeline_parts[1:])
        return module, rules

    @asyncio.coroutine
    def resolve_module(self, runtime, module_str, logging_target_name=None):
        logging_target_name = logging_target_name or module_str
        module_names = module_str.split(SCOPE_SEPARATOR)
        return (yield from self._resolve_module_from_names(
            runtime, module_names, logging_target_name))

    @asyncio.coroutine
    def _resolve_module_from_names(self, runtime, module_names,
                                   logging_target_name):
        next_module = self._get_module_checked(module_names[0])
        for name in module_names[1:]:
            next_scope = yield from _get_scope_or_fail(
                runtime, logging_target_name, next_module)
            if name not in next_scope.modules:
                _error(logging_target_name, 'module {} not found in {}', name,
                       next_module.name)
            next_module = next_scope._get_module_checked(name)
        return next_module

    @asyncio.coroutine
    def resolve_rule(self, runtime, rule_str, logging_target_name=None):
        logging_target_name = logging_target_name or rule_str
        *module_names, rule_name = rule_str.split(SCOPE_SEPARATOR)
        scope = self
        location_str = ''
        if module_names:
            module = yield from self._resolve_module_from_names(
                runtime, module_names, logging_target_name)
            scope = yield from _get_scope_or_fail(
                runtime, logging_target_name, module)
            location_str = ' in module ' + module.name
        if rule_name not in scope.rules:
            _error(logging_target_name, 'rule {} not found{}', rule_name,
                   location_str)
        return scope._get_rule_checked(rule_name)

    def get_modules_for_reup(self, names):
        for name in names:
            if SCOPE_SEPARATOR in name:
                raise PrintableError(
                    'Can\'t reup module "{}"; it belongs to another project.'
                    .format(name))
        return [self._get_module_checked(name) for name in names]

    def _get_module_checked(self, name):
        if name not in self.modules:
            raise PrintableError('Module "{}" doesn\'t exist.', name)
        return self.modules[name]

    def _get_rule_checked(self, name):
        if name not in self.rules:
            raise PrintableError('Rule "{}" doesn\'t exist.', name)
        return self.rules[name]


@asyncio.coroutine
def _get_scope_or_fail(runtime, logging_target_name, module):
    scope, imports = yield from module.parse_peru_file(runtime)
    if not scope:
        _error(logging_target_name, 'module {} is not a peru project',
               module.name)
    return scope


def _error(logging_target_name, text, *text_format_args):
    text = text.format(*text_format_args)
    raise PrintableError('Error in target {}: {}'.format(
        logging_target_name, text))



import asyncio
import json
import os
import textwrap

from .cache import compute_key
from .error import PrintableError
from .edit_yaml import set_module_field_in_file
from . import imports
from .plugin import plugin_fetch, plugin_get_reup_fields
from . import scope


recursion_warning = '''\
WARNING: The peru module '{}' doesn't specify the 'recursive' field,
but its contents include a peru.yaml file. Peru's behavior here changed
in version 0.4: modules with peru.yaml files are *no longer* recursive
by default. Add 'recursive: true' to the module definition to re-enable
recursive fetching.
'''


class Module:
    def __init__(self, name, type, default_rule, plugin_fields, yaml_name,
                 peru_file, recursive):
        self.name = name
        self.type = type
        self.default_rule = default_rule
        self.plugin_fields = plugin_fields
        # used by reup to edit the markup
        self.yaml_name = yaml_name
        # for recursive module definitions
        self.peru_file = peru_file
        # whether to automatically fetch recursive imports
        self.recursive = bool(recursive)
        self.recursion_specified = recursive is not None

    @asyncio.coroutine
    def _get_base_tree(self, runtime):
        override_path = runtime.get_override(self.name)
        if override_path is not None:
            # Marking overrides as used lets us print a warning when an
            # override is unused.
            runtime.mark_override_used(self.name)
            override_tree = yield from self._get_override_tree(
                runtime, override_path)
            return override_tree

        key = compute_key({
            'type': self.type,
            'plugin_fields': self.plugin_fields,
            'peru_file': self.peru_file,
        })
        # Use a lock to prevent the same module from being double fetched. The
        # lock is taken on the cache key, not the module itself, so two
        # different modules with identical fields will take the same lock and
        # avoid double fetching.
        cache_key_lock = runtime.cache_key_locks[key]
        with (yield from cache_key_lock):
            # Skip reading the cache if --no-cache is set. This is the only
            # place in the code we check that flag. Deterministic operations
            # like tree merging still get read from cache, because there's no
            # reason to redo them.
            if key in runtime.cache.keyval and not runtime.no_cache:
                return runtime.cache.keyval[key]
            with runtime.tmp_dir() as tmp_dir:
                yield from plugin_fetch(
                    runtime.get_plugin_context(), self.type,
                    self.plugin_fields, tmp_dir,
                    runtime.display.get_handle(self.name))
                tree = yield from runtime.cache.import_tree(tmp_dir)
            # Note that we still *write* to cache even when --no-cache is True.
            # That way we avoid confusing results on subsequent syncs.
            runtime.cache.keyval[key] = tree
        return tree

    @asyncio.coroutine
    def get_tree(self, runtime):
        # NOTE: While the recursion warning is in place, there is a 2x3 set of
        # states we want to keep track of. On the one side, a module either
        # does or does not have a peru.yaml file. On the other side, the
        # recursion setting for that module can be true, false, or unspecified.
        # It's the possible+unspecified state that we're interested in for
        # printing the warning.
        base_tree = yield from self._get_base_tree(runtime)
        scope, _imports = yield from self.parse_peru_file(runtime)
        recursion_possible = scope is not None
        if not recursion_possible:
            return base_tree
        # TODO: Get rid of this with 1.0, and move the self.recursive check up.
        if not self.recursion_specified:
            runtime.display.print(
                '\n'.join(textwrap.wrap(recursion_warning.format(self.name))))
        if not self.recursive:
            return base_tree
        recursive_tree = yield from imports.get_imports_tree(
            runtime, scope, _imports, base_tree=base_tree)
        return recursive_tree

    @asyncio.coroutine
    def parse_peru_file(self, runtime):
        from . import parser  # avoid circular imports
        tree = yield from self._get_base_tree(runtime)
        cache_key = compute_key({
            'key_type': 'module_peru_file',
            'input_tree': tree,
            'file_name': self.peru_file,
        })
        if cache_key in runtime.cache.keyval:
            yaml = json.loads(runtime.cache.keyval[cache_key])
        else:
            try:
                yaml_bytes = yield from runtime.cache.read_file(
                    tree, self.peru_file)
                yaml = yaml_bytes.decode('utf8')
            except FileNotFoundError:
                yaml = None
            runtime.cache.keyval[cache_key] = json.dumps(yaml)
        if yaml is None:
            # This module is not a peru project.
            return (None, None)
        prefix = self.name + scope.SCOPE_SEPARATOR
        return parser.parse_string(yaml, name_prefix=prefix)

    @asyncio.coroutine
    def reup(self, runtime):
        reup_fields = yield from plugin_get_reup_fields(
            runtime.get_plugin_context(), self.type, self.plugin_fields,
            runtime.display.get_handle(self.name))
        output_lines = []
        for field, val in reup_fields.items():
            if (field not in self.plugin_fields or
                    val != self.plugin_fields[field]):
                output_lines.append('  {}: {}'.format(field, val))
                set_module_field_in_file(
                    runtime.peru_file, self.yaml_name, field, val)
        if output_lines and not runtime.quiet:
            runtime.display.print('reup ' + self.name)
            for line in output_lines:
                runtime.display.print(line)

    @asyncio.coroutine
    def _get_override_tree(self, runtime, path):
        if not os.path.exists(path):
            raise PrintableError(
                "override path for module '{}' does not exist: {}".format(
                    self.name, path))
        if not os.path.isdir(path):
            raise PrintableError(
                "override path for module '{}' is not a directory: {}".format(
                    self.name, path))
        tree = yield from runtime.cache.import_tree(path)
        return tree



import asyncio
import atexit
import codecs
import contextlib
import io
import os
import subprocess
import sys
import traceback

from .error import PrintableError

# The default event loop on Windows doesn't support subprocesses, so we need to
# use the proactor loop. See:
# https://docs.python.org/3/library/asyncio-eventloops.html#available-event-loops
# Because the event loop is essentially a global variable, we have to set this
# at import time. Otherwise asyncio objects that get instantiated early
# (particularly Locks and Semaphores) could grab a reference to the wrong loop.
# TODO: Importing for side effects isn't very clean. Find a better way.
if os.name == 'nt':
    asyncio.set_event_loop(asyncio.ProactorEventLoop())

# We also need to make sure the event loop is explicitly closed, to avoid a bug
# in _UnixSelectorEventLoop.__del__. See http://bugs.python.org/issue23548.
atexit.register(asyncio.get_event_loop().close)


def run_task(coro):
    return asyncio.get_event_loop().run_until_complete(coro)


class GatheredExceptions(PrintableError):
    def __init__(self, exceptions, reprs):
        assert len(exceptions) > 0
        self.exceptions = []
        self.reprs = []
        for e, st in zip(exceptions, reprs):
            # Flatten in the exceptions list of any other GatheredExceptions we
            # see. (This happens, for example, if something throws inside a
            # recursive module.)
            if isinstance(e, GatheredExceptions):
                self.exceptions.extend(e.exceptions)
            else:
                self.exceptions.append(e)

            # Don't flatten the reprs. This would make us lose PrintableError
            # context. TODO: Represent context in a more structured way?
            self.reprs.append(st)

        self.message = "\n\n".join(self.reprs)


@asyncio.coroutine
def gather_coalescing_exceptions(coros, display, *, verbose):
    '''The tricky thing about running multiple coroutines in parallel is what
    we're supposed to do when one of them raises an exception. The approach
    we're using here is to catch exceptions and keep waiting for other tasks to
    finish. At the end, we reraise a GatheredExceptions error, if any
    exceptions were caught.

    Another minor detail: We also want to make sure to start coroutines in the
    order given, so that they end up appearing to the user alphabetically in
    the fancy display. Note that asyncio.gather() puts coroutines in a set
    internally, so we schedule coroutines *before* we give them to gather().
    '''

    exceptions = []
    reprs = []

    @asyncio.coroutine
    def catching_wrapper(coro):
        try:
            return (yield from coro)
        except Exception as e:
            exceptions.append(e)
            if isinstance(e, PrintableError) and not verbose:
                reprs.append(e.message)
            else:
                reprs.append(traceback.format_exc())
            return None

    # Suppress a deprecation warning in Python 3.5, while continuing to support
    # 3.3 and early 3.4 releases.
    if hasattr(asyncio, 'ensure_future'):
        schedule = getattr(asyncio, 'ensure_future')
    else:
        schedule = getattr(asyncio, 'async')

    futures = [schedule(catching_wrapper(coro)) for coro in coros]

    results = yield from asyncio.gather(*futures)

    if exceptions:
        raise GatheredExceptions(exceptions, reprs)
    else:
        return results


@asyncio.coroutine
def create_subprocess_with_handle(command, display_handle, *, shell=False, cwd,
                                  **kwargs):
    '''Writes subprocess output to a display handle as it comes in, and also
    returns a copy of it as a string. Throws if the subprocess returns an
    error. Note that cwd is a required keyword-only argument, on theory that
    peru should never start child processes "wherever I happen to be running
    right now."'''

    # We're going to get chunks of bytes from the subprocess, and it's possible
    # that one of those chunks ends in the middle of a unicode character. An
    # incremental decoder keeps those dangling bytes around until the next
    # chunk arrives, so that split characters get decoded properly. Use
    # stdout's encoding, but provide a default for the case where stdout has
    # been redirected to a StringIO. (This happens in tests.)
    encoding = sys.stdout.encoding or 'utf8'
    decoder_factory = codecs.getincrementaldecoder(encoding)
    decoder = decoder_factory(errors='replace')

    output_copy = io.StringIO()

    # Display handles are context managers. Entering and exiting the display
    # handle lets the display know when the job starts and stops.
    with display_handle:
        stdin = asyncio.subprocess.DEVNULL
        stdout = asyncio.subprocess.PIPE
        stderr = asyncio.subprocess.STDOUT
        if shell:
            proc = yield from asyncio.create_subprocess_shell(
                command, stdin=stdin, stdout=stdout, stderr=stderr, cwd=cwd,
                **kwargs)
        else:
            proc = yield from asyncio.create_subprocess_exec(
                *command, stdin=stdin, stdout=stdout, stderr=stderr, cwd=cwd,
                **kwargs)

        # Read all the output from the subprocess as its comes in.
        while True:
            outputbytes = yield from proc.stdout.read(4096)
            if not outputbytes:
                break
            outputstr = decoder.decode(outputbytes)
            outputstr_unified = _unify_newlines(outputstr)
            display_handle.write(outputstr_unified)
            output_copy.write(outputstr_unified)

        returncode = yield from proc.wait()

    if returncode != 0:
        raise subprocess.CalledProcessError(
            returncode, command, output_copy.getvalue())

    if hasattr(decoder, 'buffer'):
        # The utf8 decoder has this attribute, but some others don't.
        assert not decoder.buffer, 'decoder nonempty: ' + repr(decoder.buffer)

    return output_copy.getvalue()


def _unify_newlines(s):
    r'''Because all asyncio subprocess output is read in binary mode, we don't
    get universal newlines for free. But it's the right thing to do, because we
    do all our printing with strings in text mode, which translates "\n" back
    into the platform-appropriate line separator. So for example, "\r\n" in a
    string on Windows will become "\r\r\n" when it gets printed. This function
    ensures that all newlines are represented as "\n" internally, which solves
    that problem and also helps our tests work on Windows. Right now we only
    handle Windows, but we can expand this if there's ever another newline
    style we have to support.'''

    return s.replace('\r\n', '\n')


@asyncio.coroutine
def safe_communicate(process, input=None):
    '''Asyncio's communicate method has a bug where `communicate(input=b"")` is
    treated the same as `communicate(). That means that child processes can
    hang waiting for input, when their stdin should be closed. See
    https://bugs.python.org/issue26848. The issue is fixed upstream in
    https://github.com/python/asyncio/commit/915b6eaa30e1e3744e6f8223f996e197c1c9b91d,
    but we will probably always need this workaround for old versions.'''
    if input is not None and len(input) == 0:
        process.stdin.close()
        return (yield from process.communicate())
    else:
        return (yield from process.communicate(input))


class RaisesGatheredContainer:
    def __init__(self):
        self.exception = None


@contextlib.contextmanager
def raises_gathered(error_type):
    '''For use in tests. Many tests expect a single error to be thrown, and
    want it to be of a specific type. This is a helper method for when that
    type is inside a gathered exception.'''
    container = RaisesGatheredContainer()
    try:
        yield container
    except GatheredExceptions as e:
        # Make sure there is exactly one exception.
        if len(e.exceptions) != 1:
            raise
        inner = e.exceptions[0]
        # Make sure the exception is the right type.
        if not isinstance(inner, error_type):
            raise
        # Success.
        container.exception = inner



#! /usr/bin/env python3

# This plugin is for testing the fancy display. It prints some output without
# making any network requests. I needed it on an airplane :)

import random
import sys
import time

for i in range(1, 6):
    print(i)
    sys.stdout.flush()
    time.sleep(0.5 + random.random())



import asyncio
from collections import namedtuple
import contextlib
import os
import subprocess
import tempfile

import yaml

from .async import create_subprocess_with_handle
from . import cache
from . import compat
from .compat import makedirs
from .error import PrintableError

DEFAULT_PARALLEL_FETCH_LIMIT = 10

DEBUG_PARALLEL_COUNT = 0
DEBUG_PARALLEL_MAX = 0

PluginDefinition = namedtuple(
    'PluginDefinition',
    ['type', 'sync_exe', 'reup_exe', 'fields', 'required_fields',
     'optional_fields', 'cache_fields'])

PluginContext = namedtuple(
    'PluginContext',
    ['cwd', 'plugin_cache_root', 'parallelism_semaphore', 'plugin_cache_locks',
     'tmp_root'])


@asyncio.coroutine
def plugin_fetch(plugin_context, module_type, module_fields, dest,
                 display_handle):
    env = {'PERU_SYNC_DEST': dest}
    yield from _plugin_job(plugin_context, module_type, module_fields, 'sync',
                           env, display_handle)


@asyncio.coroutine
def plugin_get_reup_fields(plugin_context, module_type, module_fields,
                           display_handle):
    with tmp_dir(plugin_context) as output_file_dir:
        output_path = os.path.join(output_file_dir, 'reup_output')
        env = {'PERU_REUP_OUTPUT': output_path}
        yield from _plugin_job(
            plugin_context, module_type, module_fields, 'reup', env,
            display_handle)
        with open(output_path) as output_file:
            fields = yaml.safe_load(output_file) or {}

    for key, value in fields.items():
        if not isinstance(key, str):
            raise PluginModuleFieldError(
                'reup field name must be a string: {}'.format(key))
        if not isinstance(value, str):
            raise PluginModuleFieldError(
                'reup field value must be a string: {}'.format(value))

    return fields


@asyncio.coroutine
def _plugin_job(plugin_context, module_type, module_fields, command, env,
                display_handle):
    # We take several locks and other context managers in here. Using an
    # ExitStack saves us from indentation hell.
    with contextlib.ExitStack() as stack:
        definition = _get_plugin_definition(module_type, module_fields,
                                            command)
        exe = _get_plugin_exe(definition, command)

        # For Windows to run scripts with the right interpreter, we need to run
        # as a shell command, rather than exec.
        shell_command_line = subprocess.list2cmdline([exe])

        complete_env = _plugin_env(
            plugin_context, definition, module_fields, command, stack)
        complete_env.update(env)

        # Use a lock to protect the plugin cache. It would be unsafe for two
        # jobs to read/write to the same plugin cache dir at the same time. The
        # lock (and the cache dir) are both keyed off the module's "cache
        # fields" as defined by plugin.yaml. For plugins that don't define
        # cacheable fields, there is no cache dir (it's set to /dev/null) and
        # the cache lock is a no-op.
        stack.enter_context((yield from _plugin_cache_lock(
            plugin_context, definition, module_fields)))

        # Use a semaphore to limit the number of jobs that can run in parallel.
        # Most plugin fetches hit the network, and for performance reasons we
        # don't want to fire off too many network requests at once. See
        # DEFAULT_PARALLEL_FETCH_LIMIT. This also lets the user control
        # parallelism with the --jobs flag. It's important that this is the
        # last lock taken before starting a job, otherwise we might waste a job
        # slot just waiting on other locks.
        stack.enter_context((yield from plugin_context.parallelism_semaphore))

        # We use this debug counter for our parallelism tests. It's important
        # that it comes after all locks have been taken (so the job it's
        # counting is actually running).
        stack.enter_context(debug_parallel_count_context())

        try:
            yield from create_subprocess_with_handle(
                shell_command_line, display_handle, cwd=plugin_context.cwd,
                env=complete_env, shell=True)
        except subprocess.CalledProcessError as e:
            raise PluginRuntimeError(
                module_type, module_fields, e.returncode, e.output)


def _get_plugin_exe(definition, command):
    if command == 'sync':
        exe = definition.sync_exe
    elif command == 'reup':
        exe = definition.reup_exe
    else:
        raise RuntimeError('Unrecognized command name: ' + repr(command))

    if not exe:
        raise PluginPermissionsError(
            "Module type '{0}' does not support {1}.",
            definition.type,
            command)
    if not os.path.exists(exe):
        raise PluginPermissionsError('Plugin exe is missing: ' + exe)
    if not os.access(exe, os.X_OK):
        raise PluginPermissionsError('Plugin exe is not executable: ' + exe)
    return exe


def _format_module_fields(module_fields):
    return {'PERU_MODULE_{}'.format(name.upper()): value for
            name, value in module_fields.items()}


def _validate_plugin_definition(definition, module_fields):
    field_names_not_strings = [name for name in definition.fields
                               if not isinstance(name, str)]
    if field_names_not_strings:
        raise PluginModuleFieldError(
            'Metadata field names must be strings: ' +
            ', '.join(repr(name) for name in field_names_not_strings))

    missing_module_fields = definition.required_fields - module_fields.keys()
    if missing_module_fields:
        raise PluginModuleFieldError(
            'Required module field missing: ' +
            ', '.join(missing_module_fields))

    unknown_module_fields = module_fields.keys() - definition.fields
    if unknown_module_fields:
        raise PluginModuleFieldError(
            'Unknown module fields: ' + ', '.join(unknown_module_fields))


def _plugin_env(plugin_context, plugin_definition, module_fields, command,
                exit_stack):
    env = os.environ.copy()

    # First, blank out all module field vars.  This prevents the calling
    # environment from leaking in when optional fields are undefined.
    blank_module_vars = {field: '' for field in plugin_definition.fields}
    env.update(_format_module_fields(blank_module_vars))
    # Then add in the fields that are actually defined.
    env.update(_format_module_fields(module_fields))

    # Disable buffering by default in Python subprocesses. Without this,
    # plugins would usually need to do something like
    #     print(..., flush=True)
    # or else all their progress output would get held up in the stdout buffer
    # until the plugin finally exited. Plugins in other languages will need to
    # be careful about this.
    env['PYTHONUNBUFFERED'] = 'true'

    # For plugins that use the same exe for sync and reup, make the command
    # name available in the environment.
    env['PERU_PLUGIN_COMMAND'] = command

    # Create a directory for plugins' temporary files.
    env['PERU_PLUGIN_TMP'] = exit_stack.enter_context(tmp_dir(plugin_context))

    # Create a persistent cache dir for saved files, like repo clones.
    env['PERU_PLUGIN_CACHE'] = _plugin_cache_path(
        plugin_context, plugin_definition, module_fields)

    return env


@asyncio.coroutine
def _noop_lock():
    return contextlib.ExitStack()  # a no-op context manager


def _plugin_cache_lock(plugin_context, definition, module_fields):
    if not definition.cache_fields:
        # This plugin is not cacheable.
        return _noop_lock()
    key = _plugin_cache_key(definition, module_fields)
    return plugin_context.plugin_cache_locks[key]


def _plugin_cache_path(plugin_context, definition, module_fields):
    if not definition.cache_fields:
        # This plugin is not cacheable.
        return os.devnull
    key = _plugin_cache_key(definition, module_fields)
    plugin_cache = os.path.join(
        plugin_context.plugin_cache_root, definition.type, key)
    makedirs(plugin_cache)
    return plugin_cache


def _plugin_cache_key(definition, module_fields):
    assert definition.cache_fields, "Can't compute key for uncacheable type."
    return cache.compute_key({
        'type': definition.type,
        'cacheable_fields': {field: module_fields.get(field, None)
                             for field in definition.cache_fields},
    })


def _get_plugin_definition(module_type, module_fields, command):
    root = _find_plugin_dir(module_type)
    metadata_path = os.path.join(root, 'plugin.yaml')
    if not os.path.isfile(metadata_path):
        raise PluginMetadataMissingError(
            'No metadata file found for plugin at path: {}'.format(root))

    # Read the metadata document.
    with open(metadata_path) as metafile:
        metadoc = yaml.safe_load(metafile) or {}
    sync_exe = os.path.join(root, metadoc.pop('sync exe'))
    reup_exe = (None if 'reup exe' not in metadoc
                else os.path.join(root, metadoc.pop('reup exe')))
    required_fields = frozenset(metadoc.pop('required fields'))
    optional_fields = frozenset(metadoc.pop('optional fields', []))
    cache_fields = frozenset(metadoc.pop('cache fields', []))
    fields = required_fields | optional_fields
    # TODO: All of these checks need to be tested.
    if metadoc:
        raise RuntimeError('Unknown metadata in {} plugin: {}'.format(
            module_type, metadoc))
    overlap = required_fields & optional_fields
    if overlap:
        raise RuntimeError('Fields in {} are both required and optional: {}'
                           .format(module_type, overlap))
    invalid = cache_fields - fields
    if invalid:
        raise RuntimeError(
            '"cache fields" must also be either required or optional: ' +
            str(invalid))

    definition = PluginDefinition(
        module_type, sync_exe, reup_exe, fields, required_fields,
        optional_fields, cache_fields)
    _validate_plugin_definition(definition, module_fields)
    return definition


def _find_plugin_dir(module_type):
    '''Find the directory containing the plugin definition for the given type.
    Do this by searching all the paths where plugins can live for a dir that
    matches the type name.'''

    for install_dir in _get_plugin_install_dirs():
        candidate = os.path.join(install_dir, module_type)
        if os.path.isdir(candidate):
            return candidate
    else:
        raise PluginCandidateError(
            'No plugin found for `{}` module in paths:\n{}'.format(
                module_type,
                '\n'.join(_get_plugin_install_dirs())))


def _get_plugin_install_dirs():
    '''Return all the places on the filesystem where we should look for plugin
    definitions. Order is significant here: user-installed plugins should be
    searched first, followed by system-installed plugins, and last of all peru
    builtins.'''
    builtins_dir = os.path.join(compat.MODULE_ROOT, 'resources', 'plugins')
    if os.name == 'nt':
        # Windows
        local_data_dir = os.path.expandvars('%LOCALAPPDATA%')
        program_files_dir = os.path.expandvars('%PROGRAMFILES%')
        return (
            os.path.join(local_data_dir, 'peru', 'plugins'),
            os.path.join(program_files_dir, 'peru', 'plugins'),
            builtins_dir,
        )
    else:
        # non-Windows
        default_config_dir = os.path.expanduser('~/.config')
        config_dir = os.environ.get('XDG_CONFIG_HOME', default_config_dir)
        return (
            os.path.join(config_dir, 'peru', 'plugins'),
            '/usr/local/lib/peru/plugins',
            '/usr/lib/peru/plugins',
            builtins_dir,
        )


def debug_assert_clean_parallel_count():
    assert DEBUG_PARALLEL_COUNT == 0, \
        "parallel count should be 0 but it's " + str(DEBUG_PARALLEL_COUNT)


@contextlib.contextmanager
def debug_parallel_count_context():
    global DEBUG_PARALLEL_COUNT, DEBUG_PARALLEL_MAX
    DEBUG_PARALLEL_COUNT += 1
    DEBUG_PARALLEL_MAX = max(DEBUG_PARALLEL_COUNT, DEBUG_PARALLEL_MAX)
    try:
        yield
    finally:
        DEBUG_PARALLEL_COUNT -= 1


def tmp_dir(context):
    return tempfile.TemporaryDirectory(dir=context.tmp_root)


class PluginCandidateError(PrintableError):
    pass


class PluginCommandCandidateError(PrintableError):
    pass


class PluginModuleFieldError(PrintableError):
    pass


class PluginMetadataMissingError(PrintableError):
    pass


class PluginPermissionsError(PrintableError):
    pass


class PluginRuntimeError(PrintableError):
    def __init__(self, type, fields, errorcode, output):
        # Don't depend on plugins using terminating newlines.
        stripped_output = output.strip('\n')
        super().__init__(stripped_output)



import collections
import re
import sys
import textwrap
import yaml

from .error import PrintableError
from .module import Module
from .rule import Rule
from .scope import Scope


DEFAULT_PERU_FILE_NAME = 'peru.yaml'


class ParserError(PrintableError):
    pass


def parse_file(file_path, name_prefix=""):
    with open(file_path) as f:
        return parse_string(f.read(), name_prefix)


def parse_string(yaml_str, name_prefix=""):
    try:
        blob = yaml.safe_load(yaml_str)
    except yaml.scanner.ScannerError as e:
        raise PrintableError("YAML parser error:\n\n" + str(e)) from e
    if blob is None:
        blob = {}
    return _parse_toplevel(blob, name_prefix)


def _parse_toplevel(blob, name_prefix):
    modules = _extract_modules(blob, name_prefix)
    rules = _extract_named_rules(blob, name_prefix)
    imports = _extract_multimap_field(blob, 'imports')
    if blob:
        raise ParserError("Unknown toplevel fields: " +
                          ", ".join(blob.keys()))
    return Scope(modules, rules), imports


def _extract_named_rules(blob, name_prefix):
    scope = {}
    for field in list(blob.keys()):
        parts = field.split(' ')
        if len(parts) == 2 and parts[0] == "rule":
            _, name = parts
            if name in scope:
                raise ParserError('Rule "{}" already exists.'.format(name))
            inner_blob = typesafe_pop(blob, field)
            inner_blob = {} if inner_blob is None else inner_blob
            rule = _extract_rule(name_prefix + name, inner_blob)
            if inner_blob:
                raise ParserError("Unknown rule fields: " +
                                  ", ".join(inner_blob.keys()))
            scope[name] = rule
    return scope


def _extract_rule(name, blob):
    _validate_name(name)
    if 'build' in blob:
        raise ParserError(textwrap.dedent('''\
            The "build" field is no longer supported. If you need to
            untar/unzip a curl module, use the "unpack" field.'''))
    if 'files' in blob:
        raise ParserError(
            'The "files" field is no longer supported. Use "pick" instead.')
    copy = _extract_multimap_field(blob, 'copy')
    move = _extract_multimap_field(blob, 'move')
    executable = _extract_optional_list_field(blob, 'executable')
    drop = _extract_optional_list_field(blob, 'drop')
    pick = _extract_optional_list_field(blob, 'pick')
    export = typesafe_pop(blob, 'export', None)
    if not any((copy, move, executable, drop, pick, export)):
        return None
    rule = Rule(name, copy, move, executable, drop, pick, export)
    return rule


def _extract_default_rule(blob):
    return _extract_rule("<default>", blob)


def _extract_modules(blob, name_prefix):
    scope = {}
    for field in list(blob.keys()):
        parts = field.split(' ')
        if len(parts) == 3 and parts[1] == 'module':
            type, _, name = parts
            _validate_name(name)
            if name in scope:
                raise ParserError('Module "{}" already exists.'.format(name))
            inner_blob = typesafe_pop(blob, field)
            inner_blob = {} if inner_blob is None else inner_blob
            yaml_name = field
            module = _build_module(name_prefix + name, type, inner_blob,
                                   yaml_name)
            scope[name] = module
    return scope


def _build_module(name, type, blob, yaml_name):
    peru_file = typesafe_pop(blob, 'peru file', DEFAULT_PERU_FILE_NAME)
    recursive = typesafe_pop(blob, 'recursive', None)
    default_rule = _extract_default_rule(blob)
    plugin_fields = blob

    # Stringify all the plugin fields.
    for k, v in plugin_fields.items():
        if not isinstance(k, str):
            raise ParserError(
                'Module field names must be strings. Found "{}".'
                .format(repr(k)))
        if isinstance(v, bool):
            # Avoid the Python-specific True/False capitalization, to be
            # consistent with what people will usually type in YAML.
            plugin_fields[k] = "true" if v else "false"
        else:
            plugin_fields[k] = str(v)

    module = Module(name, type, default_rule, plugin_fields, yaml_name,
                    peru_file, recursive)
    return module


def _validate_name(name):
    if re.search(r"[\s:.]", name):
        raise ParserError("Invalid name: " + repr(name))
    return name


def _extract_optional_list_field(blob, name):
    '''Handle optional fields that can be either a string or a list of
    strings.'''
    value = _optional_list(typesafe_pop(blob, name, []))
    if value is None:
        raise ParserError('"{}" field must be a string or a list.'
                          .format(name))
    return value


def _extract_multimap_field(blob, name):
    '''Extracts multimap fields. Values can either be a scalar string or a list
    of strings. We need to parse both. For example:
        example:
          a: foo/
          b:
            - bar/
            - baz/'''
    message = ('"{}" field must be a map whose values are either a string or '
               'list of strings.'.format(name))
    raw_map = typesafe_pop(blob, name, {}) or {}
    if not isinstance(raw_map, dict):
        raise ParserError(message)
    # We use an `OrderedDict` to ensure that multimap fields are processed in a
    # determinist order. This prevents obscure bugs caused by subtly different
    # behavior.
    multimap = collections.OrderedDict()
    # Sort by key so that processing occurs in convenient lexographical order.
    # This keeps things deterministic in a simple way.
    for key, raw_value in sorted(raw_map.items()):
        value = _optional_list(raw_value)
        if value is None:
            raise ParserError(message)
        multimap[key] = value  # Remembers order.
    return multimap


def _optional_list(value):
    '''Convert a value that may be a scalar (str) or list into a tuple. This
    produces uniform output for fields that may supply a single value or list
    of values, like the `imports` field.'''
    if isinstance(value, str):
        return (value,)
    elif isinstance(value, list):
        return tuple(value)

    return None  # Let callers raise errors.


def typesafe_pop(d, field, default=object()):
    if not isinstance(d, dict):
        raise ParserError(
            'Error parsing peru file: {} is not a map.'.format(repr(d)))
    if default == typesafe_pop.__defaults__[0]:
        return d.pop(field)
    else:
        return d.pop(field, default)


# Code for the duplicate keys warning

DuplicatedKey = collections.namedtuple(
    'DuplicatedKey', ['key', 'first_line', 'second_line'])


def _get_line_indentation(line):
    indentation = 0
    for c in line:
        # YAML forbids tabs, so we only need to look for spaces.
        if c == ' ':
            indentation += 1
        else:
            return indentation


def _get_duplicate_keys_approximate(yaml_text):
    duplicates = []
    lines = yaml_text.split('\n')
    # Keep track of the keys that we've found, and the lines where we found
    # them, for every indentation level we've encountered.
    indent_to_keylines = collections.defaultdict(dict)
    for _line_index, line in enumerate(lines):
        line_num = _line_index + 1
        # Strip comments. This (and several other steps below) will do the
        # wrong thing for quoted keys that happen to contain the '#' character
        # in them, but since this is just for the sake of a warning, we don't
        # want to add all the code that would be needed to deal with it.
        if '#' in line:
            line = line[:line.index('#')]
        # Ignore lines that are not dictionary keys.
        if ':' not in line:
            continue
        current_indent = _get_line_indentation(line)
        # When an indented block ends, forget all the keys that were in it.
        for indent in list(indent_to_keylines.keys()):
            if indent > current_indent:
                del indent_to_keylines[indent]
        # Check if the current key is a duplicate.
        key = line.split(':')[0].strip()
        if key in indent_to_keylines[current_indent]:
            duplicates.append(DuplicatedKey(
                key, indent_to_keylines[current_indent][key], line_num))
        # Remember it either way.
        indent_to_keylines[current_indent][key] = line_num
    return duplicates


def _warn(s, *args, **kwargs):
    print(s.format(*args, **kwargs), file=sys.stderr)


def warn_duplicate_keys(file_path):
    with open(file_path) as f:
        text = f.read()
    duplicates = _get_duplicate_keys_approximate(text)
    if not duplicates:
        return
    _warn('WARNING: Duplicate keys found in {}\n'
          'These will overwrite each other:',
          file_path)
    for duplicate in duplicates:
        _warn('  "{}" on lines {} and {}', *duplicate)



import yaml


def set_module_field_in_file(yaml_file_path, module_name, field_name, new_val):
    with open(yaml_file_path) as f:
        yaml_text = f.read()
    new_yaml_text = set_module_field(yaml_text, module_name,
                                     field_name, new_val)
    with open(yaml_file_path, "w") as f:
        f.write(new_yaml_text)


def set_module_field(yaml_text, module_name, field_name, new_val):
    yaml_dict = _parse_yaml_text(yaml_text)
    bounds = _get_module_field_bounds(yaml_dict, module_name, field_name)
    quoted_val = _maybe_quote(new_val)
    if bounds:
        # field exists, modify it
        return yaml_text[:bounds[0]] + quoted_val + yaml_text[bounds[1]:]
    else:
        # field is new, hack it in
        return _append_module_field(yaml_text, yaml_dict, module_name,
                                    field_name, quoted_val)


def _maybe_quote(val):
    '''All of our values should be strings. Usually those can be passed in as
    bare words, but if they're parseable as an int or float we need to quote
    them.'''
    assert isinstance(val, str), 'We should never set non-string values.'
    needs_quoting = False
    try:
        int(val)
        needs_quoting = True
    except:
        pass
    try:
        float(val)
        needs_quoting = True
    except:
        pass
    if needs_quoting:
        return '"{}"'.format(val)
    else:
        return val


def _append_module_field(yaml_text, yaml_dict, module_name,
                         field_name, new_val):
    module_fields = yaml_dict[module_name]
    # use the last field to determine position and indentation
    assert len(module_fields) > 0, "There aren't any fields here!"
    last_key = module_fields.keys[-1]
    last_val = module_fields.vals[-1]
    indentation = " " * last_key.start_mark.column
    yaml_lines = yaml_text.split("\n")

    # We want to append the new field at the end of the module. Unfortunately,
    # the end_mark of a multi-line field is actually the first line of the next
    # toplevel dict. Check for this.
    if last_val.end_mark.column > 0:
        new_line_number = last_val.end_mark.line + 1
    else:
        new_line_number = last_val.end_mark.line
        # If the module ended with a line of whitespace, insert before that.
        prev_line = yaml_lines[new_line_number - 1]
        if prev_line == "" or prev_line.isspace():
            new_line_number -= 1

    new_line = "{}{}: {}".format(indentation, field_name, new_val)
    new_yaml_lines = (yaml_lines[:new_line_number] +
                      [new_line] +
                      yaml_lines[new_line_number:])
    return "\n".join(new_yaml_lines)


def _get_module_field_bounds(yaml_dict, module_name, field_name):
    module_fields = yaml_dict[module_name]
    if field_name not in module_fields:
        return None
    field_val = module_fields[field_name]
    return (field_val.start_mark.index, field_val.end_mark.index)


def _parse_yaml_text(yaml_text):
    events_list = list(yaml.parse(yaml_text))
    return _parse_events_list(events_list)


def _parse_events_list(events_list):
    event = events_list.pop(0)
    if (isinstance(event, yaml.StreamStartEvent) or
            isinstance(event, yaml.DocumentStartEvent)):
        ret = _parse_events_list(events_list)
        events_list.pop(-1)
        return ret
    elif (isinstance(event, yaml.ScalarEvent) or
          isinstance(event, yaml.AliasEvent) or
          isinstance(event, yaml.SequenceEndEvent) or
          isinstance(event, yaml.MappingEndEvent)):
        return event
    elif isinstance(event, yaml.SequenceStartEvent):
        contents = []
        while True:
            item = _parse_events_list(events_list)
            if isinstance(item, yaml.SequenceEndEvent):
                end_event = item
                return YamlList(event, end_event, contents)
            contents.append(item)
    elif isinstance(event, yaml.MappingStartEvent):
        keys = []
        vals = []
        while True:
            key = _parse_events_list(events_list)
            if isinstance(key, yaml.MappingEndEvent):
                end_event = key
                return YamlDict(event, end_event, keys, vals)
            keys.append(key)
            val = _parse_events_list(events_list)
            vals.append(val)
    else:
        raise RuntimeError("Unknown parse event type", event)


class YamlDict:
    def __init__(self, start_event, end_event, keys, vals):
        assert all(isinstance(key, yaml.ScalarEvent) for key in keys)
        assert len(keys) == len(vals)
        self.keys = keys
        self.key_map = {key.value: key for key in keys}
        self.vals = vals
        self.val_map = {key.value: val for key, val in zip(keys, vals)}
        self.start_event = start_event
        self.end_event = end_event
        self.start_mark = start_event.start_mark
        self.end_mark = end_event.end_mark

    def __contains__(self, key):
        return key in self.key_map

    def __getitem__(self, key):
        return self.val_map[key]

    def __iter__(self):
        return iter(self.key_map)

    def __len__(self):
        return len(self.keys)


class YamlList:
    def __init__(self, start_event, end_event, contents):
        self._contents = contents
        self.start_event = start_event
        self.end_event = end_event
        self.start_mark = start_event.start_mark
        self.end_mark = end_event.end_mark

    def __contains__(self, item):
        return item in self._contents

    def __getitem__(self, index):
        return self._contents[index]

    def __iter__(self):
        return iter(self._contents)

    def __len__(self):
        return len(self._contents)



import asyncio
import io
import re
import sys

# The display classes deal with output from subprocesses. The FancyDisplay
# gives a multi-line, real-time view of each running process that looks nice in
# the terminal. The VerboseDisplay collects output from each job and prints it
# all when the job is finished, in a way that's suitable for logs. The
# QuietDisplay prints nothing.
#
# All of the display types inherit from BaseDisplay and provide the same
# interface. Callers use get_handle() to get a display handle for each
# subprocess job that's going to run. The handle is used as a context manager
# (inside a with statement) to indicate when the job is starting and stopping,
# and all of the output from the subprocess is passed to the handle's write()
# method. There is also a print() method on the display, for output that's not
# tied to a particular job, which prints to the terminal in a way that won't
# get stomped on by FancyDisplay's redrawing.
#
# Like other errors, we handle job errors by throwing a PrintableError, which
# get caught in main. So the displays don't need to do anything special to show
# errors.

ANSI_CURSOR_UP_ONE_LINE = '\x1b[1A'
ANSI_CLEAR_LINE = '\x1b[2K'
ANSI_DISABLE_LINE_WRAP = '\x1b[?7l'
ANSI_ENABLE_LINE_WRAP = '\x1b[?7h'


class BaseDisplay:
    def __init__(self, output=None):
        self.output = output or sys.stdout
        # Every job/handle gets a unique id.
        self._next_job_id = 0
        # Output from each job is buffered.
        self.buffers = {}
        # Each job has a title, like the name of the module being fetched.
        self.titles = {}
        # We also keep track of any handles that haven't been entered yet, so
        # that the FancyDisplay can know when to finally clean up.
        self.outstanding_jobs = set()

    def get_handle(self, title):
        job_id = self._next_job_id
        self._next_job_id += 1
        self.titles[job_id] = title
        self.buffers[job_id] = io.StringIO()
        self.outstanding_jobs.add(job_id)
        return _DisplayHandle(self, job_id)

    # FancyDisplay overrides print() to avoid conflicting with redraws.
    def print(self, *args, **kwargs):
        print(*args, file=self.output, **kwargs)

    # Callbacks that get overridden by subclasses.

    def _job_started(self, job_id):
        pass

    def _job_written(self, job_id, string):
        pass

    def _job_finished(self, job_id):
        pass

    # Callbacks for handles.

    def _handle_start(self, job_id):
        self._job_started(job_id)

    def _handle_write(self, job_id, string):
        self.buffers[job_id].write(string)
        self._job_written(job_id, string)

    def _handle_finish(self, job_id):
        self.outstanding_jobs.remove(job_id)
        self._job_finished(job_id)


class QuietDisplay(BaseDisplay):
    '''Prints nothing.'''
    pass


class VerboseDisplay(BaseDisplay):
    '''Waits until jobs are finished and then prints all of their output at
    once, to make sure jobs don't get interleaved. We use '===' as a delimiter
    to try to separate jobs from one another, and from other output.'''
    def _job_started(self, job_id):
        print('===', 'started', self.titles[job_id], '===', file=self.output)

    def _job_finished(self, job_id):
        print('===', 'finished', self.titles[job_id], '===', file=self.output)
        outputstr = self.buffers[job_id].getvalue()
        if outputstr:
            self.output.write(outputstr)
            print('===', file=self.output)


class FancyDisplay(BaseDisplay):
    '''Prints a multi-line, real-time display of all the latest output lines
    from each job.'''
    def __init__(self, *args):
        super().__init__(*args)
        # Every time we draw we need to erase the lines that were printed
        # before. This keeps track of that number. Note that we split output on
        # newlines and use no-wrap control codes in the terminal, so we only
        # need to count the number of jobs drawn.
        self._lines_printed = 0
        # This is the list of all active jobs. There's no guarantee that jobs
        # start in any particular order, so this list also helps us keep the
        # order stable.
        self._job_slots = []
        # The last line output from each job. This is what gets drawn.
        self._output_lines = {}
        # Lines that need to be printed above the display. This has to happen
        # during the next draw, right after the display is cleared.
        self._to_print = []
        # To avoid flicker, we draw on a short timeout instead of every time we
        # receive output. When this asyncio handle is set, it means a draw is
        # already pending.
        self._draw_later_handle = None

    def print(self, *args, **kwargs):
        output = io.StringIO()
        print(*args, file=output, **kwargs)
        self._to_print.append(output.getvalue())
        # If we use _draw_later, the program might exit before the draw timer
        # fires. Drawing right now ensures that output never gets dropped.
        self._draw()

    def _clear(self, *, flush=True):
        # Erase everything we printed before.
        for i in range(self._lines_printed):
            self.output.write(ANSI_CURSOR_UP_ONE_LINE)
            self.output.write(ANSI_CLEAR_LINE)
        if flush:
            self.output.flush()
        self._lines_printed = 0

    def _draw(self):
        self._cancel_draw_later()

        self._clear(flush=False)  # We will flush below.

        # If we have any lines from print(), print them now. They will end up
        # above the display like regular output.
        for string in self._to_print:
            self.output.write(string)
        self._to_print.clear()

        # Redraw all the jobs.
        self.output.write(ANSI_DISABLE_LINE_WRAP)
        for slot, job_id in enumerate(self._job_slots):
            # Fancy unicode box characters in the left column.
            if slot == 0:
                self.output.write('┌' if len(self._job_slots) > 1 else '╶')
            elif slot < len(self._job_slots) - 1:
                self.output.write('├')
            else:
                self.output.write('└')
            self.output.write(' ')
            self.output.write(self.titles[job_id])
            self.output.write(': ')
            self.output.write(self._output_lines[job_id])
            # Some terminals keep overwriting the last character in no-wrap
            # mode. Make the trailing character a space.
            self.output.write(' ')
            self.output.write('\n')
            self._lines_printed += 1
        self.output.write(ANSI_ENABLE_LINE_WRAP)
        self.output.flush()

    def _draw_later(self):
        if self._draw_later_handle:
            # There is already a draw pending.
            return
        self._draw_later_handle = asyncio.get_event_loop().call_later(
            0.1, self._draw)

    def _cancel_draw_later(self):
        if self._draw_later_handle:
            self._draw_later_handle.cancel()
            self._draw_later_handle = None

    def _job_started(self, job_id):
        self._job_slots.append(job_id)
        self._output_lines[job_id] = ''
        self._draw_later()

    def _job_written(self, job_id, string):
        # We need to split output on newlines. Some programs (git) also use
        # carriage return to redraw a line, so we split on that too.
        any_newlines = '(?:\n|\r)+'  # (?: is non-capturing, for split()
        lines = [line.strip() for line in re.split(any_newlines, string)]

        # NB: We don't make any attempt here to join lines that might span
        # multiple write() calls. `async.create_subprocess_with_handle()` reads
        # output in 4096 byte chunks, so this isn't likely, but it's possible.
        for line in lines:
            # Ignore empty lines, both from the job and from re.split().
            if line:
                self._output_lines[job_id] = line
        self._draw_later()

    def _job_finished(self, job_id):
        self._job_slots.remove(job_id)
        if not self.outstanding_jobs:
            # If the last job is finished, the event loop might be about to
            # stop. Clear the terminal right now, because _draw_later might
            # never run.
            self._draw()
        else:
            # If there are pending jobs, don't clear the display immediately.
            # This avoid flickering between jobs when only one job is running
            # at a time (-j1).
            self._draw_later()


class _DisplayHandle:
    def __init__(self, display, job_id):
        self._display = display
        self._job_id = job_id
        self._opened = False
        self._closed = False

    def write(self, string):
        assert self._opened and not self._closed
        self._display._handle_write(self._job_id, string)

    # Context manager interface. We're extra careful to make sure that the
    # handle is only written to inside a with statment, and only used once.
    def __enter__(self):
        assert not self._opened and not self._closed
        self._opened = True
        self._display._handle_start(self._job_id)
        return self

    def __exit__(self, *args):
        assert self._opened and not self._closed
        self._display._handle_finish(self._job_id)
        self._job_id = None
        self._closed = True



#! /usr/bin/env python3

import configparser
import hashlib
import os
import subprocess
import sys

URL = os.environ['PERU_MODULE_URL']
REV = os.environ['PERU_MODULE_REV'] or 'master'
REUP = os.environ['PERU_MODULE_REUP'] or 'master'

# Because peru gives each plugin a unique cache dir based on its cacheable
# fields (in this case, url) we could clone directly into cache_root. However,
# because the git plugin needs to handle git submodules as well, it still has
# to separate things out by repo url.
CACHE_ROOT = os.environ['PERU_PLUGIN_CACHE']


def git(*args, git_dir=None, capture_output=False):
    # Avoid forgetting this arg.
    assert git_dir is None or os.path.isdir(git_dir)

    command = ['git']
    if git_dir:
        command.append('--git-dir={0}'.format(git_dir))
    command.extend(args)

    stdout = subprocess.PIPE if capture_output else None
    # Always let stderr print to the caller.
    process = subprocess.Popen(
        command,
        stdin=subprocess.DEVNULL,
        stdout=stdout,
        universal_newlines=True)
    output, _ = process.communicate()
    if process.returncode != 0:
        sys.exit(1)

    return output


def has_clone(url):
    return os.path.exists(repo_cache_path(url))


def clone_if_needed(url):
    repo_path = repo_cache_path(url)
    if not has_clone(url):
        git('clone', '--mirror', '--progress', url, repo_path)
    return repo_path


def repo_cache_path(url):
    # If we just concatenate the escaped repo URL into the path, we start to
    # run up against the 260-character path limit on Windows.
    url_hash = hashlib.sha1(url.encode()).hexdigest()
    return os.path.join(CACHE_ROOT, url_hash)


def clone_and_maybe_print(url):
    if not has_clone(url):
        print('git clone ' + url)
    return clone_if_needed(url)


def git_fetch(url, repo_path):
    print('git fetch ' + url)
    git('fetch', '--prune', git_dir=repo_path)


def already_has_rev(repo, rev):
    try:
        # Make sure the rev exists.
        git('cat-file', '-e', rev, git_dir=repo)
        # Get the hash for the rev.
        output = git('rev-parse', rev, git_dir=repo, capture_output=True)
    except:
        return False

    # Only return True for revs that are absolute hashes.
    # We could consider treating tags the way, but...
    # 1) Tags actually can change.
    # 2) It's not clear at a glance if something is a branch or a tag.
    # Keep it simple.
    return output.strip() == rev


def checkout_tree(url, rev, dest):
    repo_path = clone_and_maybe_print(url)
    if not already_has_rev(repo_path, rev):
        git_fetch(url, repo_path)
    # If we just use `git checkout rev -- .` here, we get an error when rev is
    # an empty commit.
    git('--work-tree=' + dest, 'read-tree', rev, git_dir=repo_path)
    git('--work-tree=' + dest, 'checkout-index', '--all', git_dir=repo_path)
    checkout_submodules(repo_path, rev, dest)


def checkout_submodules(repo_path, rev, work_tree):
    if os.environ['PERU_MODULE_SUBMODULES'] == 'false':
        return

    gitmodules = os.path.join(work_tree, '.gitmodules')
    if not os.path.exists(gitmodules):
        return

    parser = configparser.ConfigParser()
    parser.read(gitmodules)
    for section in parser.sections():
        sub_relative_path = parser[section]['path']
        sub_full_path = os.path.join(work_tree, sub_relative_path)
        sub_url = parser[section]['url']
        ls_tree = git('ls-tree', rev, sub_relative_path,
                      git_dir=repo_path, capture_output=True)
        # Normally when you run `git submodule add ...`, git puts two things in
        # your repo: an entry in .gitmodules, and a commit object at the
        # appropriate path inside your repo. However, it's possible for those
        # two to get out of sync, especially if you use mv/rm on a directory
        # followed by `git add`, instead of the smarter `git mv`/`git rm`. If
        # we run into one of these missing submodules, just skip it.
        if len(ls_tree.strip()) == 0:
            print('WARNING: sudmodule ' + sub_relative_path +
                  ' is configured in .gitmodules, but missing in the repo')
            continue
        sub_rev = ls_tree.split()[2]
        checkout_tree(sub_url, sub_rev, sub_full_path)


def plugin_sync():
    checkout_tree(URL, REV, os.environ['PERU_SYNC_DEST'])


def plugin_reup():
    reup_output = os.environ['PERU_REUP_OUTPUT']
    repo_path = clone_if_needed(URL)
    git_fetch(URL, repo_path)
    output = git('rev-parse', REUP, git_dir=repo_path, capture_output=True)
    with open(reup_output, 'w') as out_file:
        print('rev:', output.strip(), file=out_file)

command = os.environ['PERU_PLUGIN_COMMAND']
if command == 'sync':
    plugin_sync()
elif command == 'reup':
    plugin_reup()
else:
    raise RuntimeError('Unknown command: ' + repr(command))



#! /usr/bin/env python3

import os
import subprocess
import sys
import textwrap

CACHE_PATH = os.environ['PERU_PLUGIN_CACHE']
URL = os.environ['PERU_MODULE_URL']
REV = os.environ['PERU_MODULE_REV'] or 'default'
REUP = os.environ['PERU_MODULE_REUP'] or 'default'


def hg(*args, hg_dir=None, capture_output=False):
    # Avoid forgetting this arg.
    assert hg_dir is None or os.path.isdir(hg_dir)

    command = ['hg']
    if hg_dir:
        command.append('--repository')
        command.append(hg_dir)
    command.extend(args)

    stdout = subprocess.PIPE if capture_output else None
    # Always let stderr print to the caller.
    process = subprocess.Popen(command, stdin=subprocess.DEVNULL,
                               stdout=stdout, universal_newlines=True)
    output, _ = process.communicate()
    if process.returncode != 0:
        sys.exit(1)

    return output


def clone_if_needed(url, verbose=False):
    if not os.path.exists(os.path.join(CACHE_PATH, '.hg')):
        if verbose:
            print('hg clone', url)
        hg('clone', '--noupdate', url, CACHE_PATH)
        configure(CACHE_PATH)


def configure(repo_path):
    # Set configs needed for cached repos.
    hgrc_path = os.path.join(repo_path, '.hg', 'hgrc')
    with open(hgrc_path, 'a') as f:
        f.write(textwrap.dedent('''\
            [ui]
            # prevent 'hg archive' from creating '.hg_archival.txt' files.
            archivemeta = false
            '''))


def hg_pull(url, repo_path):
    print('hg pull', url)
    hg('pull', hg_dir=repo_path)


def already_has_rev(repo, rev):
    try:
        output = hg('identify', '--debug', '--rev', rev, hg_dir=repo,
                    capture_output=True)
    except:
        return False

    # Only return True for revs that are absolute hashes.
    # We could consider treating tags the way, but...
    # 1) Tags actually can change.
    # 2) It's not clear at a glance whether something is a branch or a tag.
    # Keep it simple.
    return output.split()[0] == rev


def plugin_sync():
    dest = os.environ['PERU_SYNC_DEST']
    clone_if_needed(URL, verbose=True)
    if not already_has_rev(CACHE_PATH, REV):
        hg_pull(URL, CACHE_PATH)
    # TODO: Should this handle subrepos?
    hg('archive', '--type', 'files', '--rev', REV, dest, hg_dir=CACHE_PATH)


def plugin_reup():
    reup_output = os.environ['PERU_REUP_OUTPUT']

    clone_if_needed(URL, CACHE_PATH)
    hg_pull(URL, CACHE_PATH)
    output = hg('identify', '--debug', '--rev', REUP, hg_dir=CACHE_PATH,
                capture_output=True)

    with open(reup_output, 'w') as output_file:
        print('rev:', output.split()[0], file=output_file)


command = os.environ['PERU_PLUGIN_COMMAND']
if command == 'sync':
    plugin_sync()
elif command == 'reup':
    plugin_reup()
else:
    raise RuntimeError('Unknown command: ' + repr(command))



#! /usr/bin/env python3

import hashlib
import os
import pathlib
import re
import stat
import sys
import tarfile
from urllib.parse import urlsplit
import urllib.request
import zipfile


def get_request_filename(request):
    '''Figure out the filename for an HTTP download.'''
    # Check to see if a filename is specified in the HTTP headers.
    if 'Content-Disposition' in request.info():
        disposition = request.info()['Content-Disposition']
        pieces = re.split('\s*;\s*', disposition)
        for piece in pieces:
            if piece.startswith('filename='):
                filename = piece[len('filename='):]
                # Strip exactly one " from each end.
                if filename.startswith('"'):
                    filename = filename[1:]
                if filename.endswith('"'):
                    filename = filename[:-1]
                # Interpret backslashed quotes.
                filename = filename.replace('\\"', '"')
                return filename
    # If no filename was specified, pick a reasonable default.
    return os.path.basename(urlsplit(request.url).path) or 'index.html'


def format_bytes(num_bytes):
    for threshold, unit in ((10**9, 'GB'), (10**6, 'MB'), (10**3, 'KB')):
        if num_bytes >= threshold:
            # Truncate floats instead of rounding.
            float_str = str(num_bytes / threshold)
            decimal_index = float_str.index('.')
            truncated_float = float_str[:decimal_index+2]
            return truncated_float + unit
    return '{}B'.format(num_bytes)


def download_file(request, output_file, stdout=sys.stdout):
    digest = hashlib.sha1()
    file_size_str = request.info().get('Content-Length')
    file_size = int(file_size_str) if file_size_str is not None else None
    bytes_read = 0
    while True:
        buf = request.read(4096)
        if not buf:
            break
        digest.update(buf)
        if output_file:
            output_file.write(buf)
        bytes_read += len(buf)
        percentage = ''
        kb_downloaded = format_bytes(bytes_read)
        total_kb = ''
        if file_size:
            percentage = ' {}%'.format(round(100 * bytes_read / file_size))
            total_kb = '/' + format_bytes(file_size)
        print('downloaded{} {}{}'.format(percentage, kb_downloaded, total_kb),
              file=stdout)
    return digest.hexdigest()


def plugin_sync(url, sha1):
    unpack = os.environ['PERU_MODULE_UNPACK']
    dest = os.environ['PERU_SYNC_DEST']
    if unpack:
        # Download to the tmp dir for later unpacking.
        download_dir = os.environ['PERU_PLUGIN_TMP']
    else:
        # Download directly to the destination dir.
        download_dir = dest

    with urllib.request.urlopen(url) as request:
        filename = os.environ['PERU_MODULE_FILENAME']
        if not filename:
            filename = get_request_filename(request)
        full_filepath = os.path.join(download_dir, filename)
        with open(full_filepath, 'wb') as output_file:
            digest = download_file(request, output_file)

    if sha1 and digest != sha1:
        print('Bad checksum!\n     url: {}\nexpected: {}\n  actual: {}'
              .format(url, sha1, digest), file=sys.stderr)
        sys.exit(1)

    try:
        if unpack == 'tar':
            extract_tar(full_filepath, dest)
        elif unpack == 'zip':
            extract_zip(full_filepath, dest)
        elif unpack:
            print('Unknown value for "unpack":', unpack, file=sys.stderr)
            sys.exit(1)
    except EvilArchiveError as e:
        print(e.message, file=sys.stderr)
        sys.exit(1)


def extract_tar(archive_path, dest):
    with tarfile.open(archive_path) as t:
        validate_filenames(info.path for info in t.getmembers())
        t.extractall(dest)


def extract_zip(archive_path, dest):
    with zipfile.ZipFile(archive_path) as z:
        validate_filenames(z.namelist())
        z.extractall(dest)
        # Set file permissions. Tar does this by default, but with zip we need
        # to do it ourselves.
        for info in z.filelist:
            if not info.filename.endswith('/'):
                # This is how to get file permissions out of a zip archive,
                # according to http://stackoverflow.com/q/434641/823869 and
                # http://bugs.python.org/file34873/issue15795_cleaned.patch.
                mode = (info.external_attr >> 16) & 0o777
                # Don't copy the whole mode, just set the executable bit. Two
                # reasons for this. 1) This is all going to end up in a git
                # tree, which only records the executable bit anyway. 2) Zip's
                # support for Unix file modes is nonstandard, so the mode field
                # is often zero and could be garbage. Mistakenly setting a file
                # executable isn't a big deal, but e.g. removing read
                # permissions would cause an error.
                if mode & stat.S_IXUSR:
                    os.chmod(os.path.join(dest, info.filename), 0o755)


def validate_filenames(names):
    for name in names:
        path = pathlib.PurePosixPath(name)
        if path.is_absolute() or '..' in path.parts:
            raise EvilArchiveError('Illegal path in archive: ' + name)


class EvilArchiveError(RuntimeError):
    def __init__(self, message):
        self.message = message


def plugin_reup(url, sha1):
    reup_output = os.environ['PERU_REUP_OUTPUT']
    with urllib.request.urlopen(url) as request:
        digest = download_file(request, None)
    with open(reup_output, 'w') as output_file:
        print('sha1:', digest, file=output_file)


def main():
    url = os.environ['PERU_MODULE_URL']
    sha1 = os.environ['PERU_MODULE_SHA1']
    command = os.environ['PERU_PLUGIN_COMMAND']
    if command == 'sync':
        plugin_sync(url, sha1)
    elif command == 'reup':
        plugin_reup(url, sha1)
    else:
        raise RuntimeError('unknown command: ' + repr(command))


if __name__ == '__main__':
    main()



import os
import setuptools
import sys

# Written according to the docs at
# https://packaging.python.org/en/latest/distributing.html

project_root = os.path.dirname(__file__)
peru_sh_path = os.path.join(project_root, 'peru.sh')
module_root = os.path.join(project_root, 'peru')
version_file = os.path.join(module_root, 'VERSION')


def get_version():
    with open(version_file) as f:
        return f.read().strip()


def get_all_resources_filepaths():
    resources_paths = ['VERSION']
    resources_dir = os.path.join(module_root, 'resources')
    for dirpath, dirnames, filenames in os.walk(resources_dir):
        relpaths = [os.path.relpath(os.path.join(dirpath, f),
                                    start=module_root)
                    for f in filenames]
        resources_paths.extend(relpaths)
    return resources_paths


def get_install_requires():
    dependencies = ['docopt', 'PyYAML']
    # Python 3.3 needs extra libs that aren't installed by default.
    if sys.version_info < (3, 3):
        raise RuntimeError('The minimum supported Python version is 3.3.')
    elif (3, 3) <= sys.version_info < (3, 4):
        dependencies.extend(['asyncio', 'pathlib'])
    return dependencies


setuptools.setup(
    name='peru',
    description='A tool for fetching code',
    version=get_version(),
    url='https://github.com/buildinspace/peru',
    author="Jack O'Connor <oconnor663@gmail.com>, "
           "Sean Olson <olson.sean.k@gmail.com>",
    license='MIT',
    packages=['peru'],
    package_data={'peru': get_all_resources_filepaths()},
    entry_points={
        'console_scripts': [
            'peru=peru.main:main',
        ]
    },
    install_requires=get_install_requires(),
)



import contextlib
import os
import shutil
import tempfile

from . import compat


class KeyVal:
    '''A generic way to store key-value pairs on disk. Just creates files in a
    folder whose names are the keys and whose contents are the values.'''

    def __init__(self, root, tmp_dir):
        self._root = root
        self._tmp_dir = tmp_dir
        compat.makedirs(root)
        compat.makedirs(tmp_dir)

    def __getitem__(self, key):
        with open(self._path(key)) as f:
            return f.read()

    def __setitem__(self, key, val):
        # Write to a tmp file first, to avoid partial reads.
        tmp_path = self._tmp_file()
        with open(tmp_path, "w") as f:
            f.write(val)
        shutil.move(tmp_path, self._path(key))

    def __delitem__(self, key):
        if os.path.exists(self._path(key)):
            os.remove(self._path(key))

    def __contains__(self, key):
        return os.path.isfile(self._path(key))

    def __iter__(self):
        return iter(os.listdir(self._root))

    def __len__(self):
        return len(os.listdir(self._root))

    def _path(self, key):
        return os.path.join(self._root, key)

    def _tmp_file(self):
        fd, path = tempfile.mkstemp(dir=self._tmp_dir)
        os.close(fd)
        return path

    @contextlib.contextmanager
    def tmp_dir_context(self):
        try:
            path = tempfile.mkdtemp(dir=self._tmp_dir)
            yield path
        finally:
            shutil.rmtree(path, ignore_errors=True)



#! /usr/bin/env python3

import asyncio
import collections
import os
import sys
import tempfile

import docopt

from . import async
from . import compat
from .error import PrintableError
from . import imports
from . import parser
from .runtime import Runtime


__doc__ = '''\
Usage:
    peru [-hqv] [--file=<file>] [--sync-dir=<dir>] [--state-dir=<dir>]
         [--cache-dir=<dir>] [--file-basename=<name>] <command> [<args>...]
    peru [--help|--version]

Commands:
    sync      fetch imports and copy them to your project
    reup      update revision information for your modules
    clean     delete imports from your project
    copy      copy files directly from a module to somewhere else
    override  substitute a local directory for the contents of a module
    help      show help for subcommands, same as -h/--help

Options:
    -h --help             so much help
    -q --quiet            don't print anything
    -v --verbose          print everything

    --file=<file>
        The project file to use instead of 'peru.yaml'. This must be used
        together with --sync-dir.
    --sync-dir=<dir>
        The root directory for your imports, instead of the directory
        containing 'peru.yaml'. This must be used together with --file.
    --state-dir=<dir>
        The directory where peru keeps all of its metadata, including the cache
        and the current imports tree. Defaults to '.peru' next to 'peru.yaml'.
    --cache-dir=<dir>
        The directory for caching all the files peru fetches. Defaults to
        '.peru/cache', or $PERU_CACHE_DIR if it's defined.
    --file-basename=<name>
        An alternative filename (not a path) for 'peru.yaml'. As usual, peru
        will search the current dir and its parents for this file, and import
        paths will be relative to it. Incompatible with --file.
'''


def peru_command(name, doc):
    def decorator(f):
        coro = asyncio.coroutine(f)
        COMMAND_FNS[name] = coro
        COMMAND_DOCS[name] = doc
        return coro
    return decorator

COMMAND_FNS = {}
COMMAND_DOCS = {}


@peru_command('sync', '''\
Usage:
    peru sync [-fhqv] [-j N] [--no-cache] [--no-overrides]

Writes your imports to the sync directory. By default, this is the
directory that contains your peru.yaml file. Peru is normally careful
not to overwrite pre-existing or modified files, and if it detects any
then it writes nothing and reports an error. Use the --force flag if you
want sync to overwrite existing files.

Options:
    -f --force      overwrite existing or changed files
    -h --help       explain these confusing flags
    -j N --jobs N   max number of parallel fetches
    --no-cache      force modules without exact revs to refetch
    --no-overrides  suppress any `peru override` settings
    -q --quiet      don't print anything
    -v --verbose    print everything
''')
def do_sync(params):
    params.runtime.print_overrides()
    yield from imports.checkout(
        params.runtime, params.scope, params.imports, params.runtime.sync_dir)
    params.runtime.warn_unused_overrides()


@peru_command('reup', '''\
Usage:
    peru reup [<modules>...] [-fhqv] [-j N] [--no-cache] [--no-overrides]
              [--no-sync]

Updates each module in your peru.yaml file with the latest revision
information from its source. For git, hg, and svn modules, this is the
`rev` field, and for a curl module it's the `sha1` field. Peru will
either add the field for you, or update it if it's already there. To
update specific modules instead of everything, pass their names as
positional arguments. Peru does a sync after the reup is done, though
you can prevent that with --no-sync.

Options:
    -f --force      for `peru sync`
    -h --help       what is even happening here?
    --no-cache      for `peru sync`
    --no-overrides  for `peru sync`
    --no-sync       skip the sync at the end
    -j N --jobs N   max number of parallel fetches
    -q --quiet      don't print anything
    -v --verbose    print everything
''')
def do_reup(params):
    names = params.args['<modules>']
    if not names:
        modules = params.scope.modules.values()
    else:
        modules = params.scope.get_modules_for_reup(names)
    futures = [module.reup(params.runtime) for module in modules]
    yield from async.gather_coalescing_exceptions(
        futures,
        params.runtime.display,
        verbose=params.runtime.verbose)
    if not params.args['--no-sync']:
        # Do an automatic sync. Reparse peru.yaml to get the new revs.
        new_scope, new_imports = parser.parse_file(params.runtime.peru_file)
        new_params = params._replace(scope=new_scope, imports=new_imports)
        yield from do_sync(new_params)


@peru_command('clean', '''\
Usage:
    peru clean [-fhqv]

Removes any files previously written by sync. As with sync, peru is
cautious about removing files that you have changed. Use --force to
clean changed files.

Options:
    -f --force     allow cleaning modified files
    -h --help      what were they thinking?
    -q --quiet     don't print anything
    -v --verbose   print everything
''')
def do_clean(params):
    yield from imports.checkout(
        params.runtime, params.scope, {}, params.runtime.sync_dir)


@peru_command('copy', '''\
Usage:
    peru copy <target> [<dest>] [-fhqv] [-j N] [--no-cache] [--no-overrides]
    peru copy --help

Writes the contents of a target to a temp dir, or to a destination that
you specify. A target is anything that you can import, so it can be just
a module (foo), a module followed by named rules (foo|bar|baz), or even
a module defined within another module (foo.bing).

Options:
    -f --force      overwrite existing files
    -h --help       is anyone even listening?
    -j N --jobs N   max number of parallel fetches
    --no-cache      force modules without exact revs to refetch
    --no-overrides  suppress any `peru override` settings
    -q --quiet      don't print anything
    -v --verbose    print everything
''')
def do_copy(params):
    params.runtime.print_overrides()
    if not params.args['<dest>']:
        dest = tempfile.mkdtemp(prefix='peru_copy_')
    else:
        dest = params.args['<dest>']
    tree = yield from imports.get_tree(
        params.runtime, params.scope, params.args['<target>'])
    yield from params.runtime.cache.export_tree(
        tree, dest, force=params.runtime.force)
    if not params.args['<dest>']:
        print(dest)


@peru_command('override', '''\
Usage:
    peru override [list]
    peru override add <module> <path>
    peru override delete <module>
    peru override --help

Adding an override tells peru to use the contents of a given directory
in place of the actual contents of a module. So for example, if your
project normally fetches `foo` from GitHub, but you want to test it with
some changes to `foo` that don't exist upstream, you can override the
`foo` module to with your clone. Then the next time you sync, peru will
use the files from your clone.

Note that override directories are independent of the type of your
module. Peru doesn't care whether you override your git module with an
actual git clone, or just a directory full of files. It simply copies
what's there. Module fields (including `rev`) have no effect while a
module is overridden.

Options:
    -h --help  (>'-')> <('-'<) ^('-')^
''')
def do_override(params):
    if params.args['add']:
        name = params.args['<module>']
        path = params.args['<path>']
        params.runtime.set_override(name, path)
    elif params.args['delete']:
        key = params.args['<module>']
        del params.runtime.overrides[key]
    else:
        for module in sorted(params.runtime.overrides):
            print('{}: {}'.format(module, params.runtime.get_override(module)))


def get_version():
    version_file = os.path.join(compat.MODULE_ROOT, 'VERSION')
    with open(version_file) as f:
        return f.read().strip()


def print_red(*args, **kwargs):
    if compat.is_fancy_terminal():
        sys.stdout.write('\x1b[31m')
    print(*args, **kwargs)
    if compat.is_fancy_terminal():
        sys.stdout.write('\x1b[39m')


def maybe_print_help_and_return(args):
    # `peru --version`
    if args['--version']:
        print(get_version())
        return 0

    help = args['--help']
    command = args['<command>']
    if command == "help":
        help = True
        help_args = args['<args>']
        command = help_args[0] if help_args else None

    # no explicit command, just print toplevel help
    if command is None:
        print(__doc__, end='')
        return 0

    # bad command, or help for a bad command
    if command not in COMMAND_DOCS:
        print(__doc__, end='', file=sys.stderr)
        return 1

    # help for a specific command that actually exists
    if help:
        doc = COMMAND_DOCS.get(command, __doc__)
        print(doc, end='')
        return 0

    # otherwise help is not called for
    return None


def merged_args_dicts(global_args, subcommand_args):
    '''We deal with docopt args from the toplevel peru parse and the subcommand
    parse. We don't want False values for a flag in the subcommand to override
    True values if that flag was given at the top level. This function
    specifically handles that case.'''
    merged = global_args.copy()
    for key, val in subcommand_args.items():
        if key not in merged:
            merged[key] = val
        elif type(merged[key]) is type(val) is bool:
            merged[key] = merged[key] or val
        else:
            raise RuntimeError("Unmergable args.")
    return merged


def docopt_parse_args(argv):
    args = docopt.docopt(__doc__, argv, help=False, options_first=True)
    command = args['<command>']
    # Skip further parsing for cases like `peru badcommand` (because there is
    # no docopt), `peru help <cmd>` (because help is a fake command also with
    # no docopt), and `peru --help copy` (because e.g. the copy subcommand
    # does not support calls with no args).
    if command in COMMAND_DOCS and not args['--help']:
        command_doc = COMMAND_DOCS[command]
        command_argv = [command] + args['<args>']
        command_args = docopt.docopt(command_doc, command_argv, help=False)
        args = merged_args_dicts(args, command_args)
    return args


CommandParams = collections.namedtuple(
    'CommandParams', ['args', 'runtime', 'scope', 'imports'])


def force_utf8_in_ascii_mode_hack():
    '''In systems without a UTF8 locale configured, Python will default to
    ASCII mode for stdout and stderr. This causes our fancy display to fail
    with encoding errors. In particular, you run into this if you try to run
    peru inside of Docker. This is a hack to force emitting UTF8 in that case.
    Hopefully it doesn't break anything important.'''
    if sys.stdout.encoding == 'ANSI_X3.4-1968':
        sys.stdout = open(sys.stdout.fileno(), mode='w', encoding='utf8',
                          buffering=1)
        sys.stderr = open(sys.stderr.fileno(), mode='w', encoding='utf8',
                          buffering=1)


def main(*, argv=None, env=None, nocatch=False):
    force_utf8_in_ascii_mode_hack()

    if argv is None:
        argv = sys.argv[1:]
    if env is None:
        env = os.environ.copy()

    args = docopt_parse_args(argv)
    command = args['<command>']

    ret = maybe_print_help_and_return(args)
    if ret is not None:
        return ret

    try:
        runtime = async.run_task(Runtime(args, env))
        if not args['--quiet']:
            parser.warn_duplicate_keys(runtime.peru_file)
        scope, imports = parser.parse_file(runtime.peru_file)
        params = CommandParams(args, runtime, scope, imports)
        command_fn = COMMAND_FNS[command]
        async.run_task(command_fn(params))
    except PrintableError as e:
        if args['--verbose'] or nocatch:
            # Just allow the stacktrace to print if verbose, or in testing.
            raise
        print_red(e.message, end='' if e.message.endswith('\n') else '\n')
        return 1


if __name__ == '__main__':
    main()



from contextlib import contextmanager
from textwrap import indent


class PrintableError(Exception):
    def __init__(self, message, *args, **kwargs):
        self.message = message.format(*args, **kwargs)

    def __str__(self):
        return self.message

    def add_context(self, context):
        # TODO: Something more structured?
        self.message = 'In {}:\n{}'.format(context, indent(self.message, '  '))


@contextmanager
def error_context(context):
    try:
        yield
    except PrintableError as e:
        e.add_context(context)
        raise



import asyncio
import contextlib
import collections
import hashlib
import json
import os
import pathlib
import re
import textwrap

from .async import safe_communicate
from .compat import makedirs
from .error import PrintableError
from .keyval import KeyVal


# git output modes
TEXT_MODE = object()
BINARY_MODE = object()

# for tests
DEBUG_GIT_COMMAND_COUNT = 0


def compute_key(data):
    # To hash this dictionary of fields, serialize it as a JSON string, and
    # take the SHA1 of that string. Dictionary key order is unspecified, so
    # "sort_keys" keeps our hash stable. Specifying separators makes the
    # JSON slightly more compact, and protects us against changes in the
    # default.  "ensure_ascii" defaults to true, so specifying it just
    # protects us from changes in the default.
    json_representation = json.dumps(
        data, sort_keys=True, ensure_ascii=True, separators=(',', ':'))
    sha1 = hashlib.sha1()
    sha1.update(json_representation.encode("utf8"))
    return sha1.hexdigest()


class GitSession:
    '''All of our git operations will share the same repo, but we don't want
    them to share the same index file. That's for two reasons:
      1) We want to be able to run multiple operations in parallel that write
         to the index file.
      2) We want to be able to save the index file corresponding to the last
         imports, and guarantee that nothing will touch it.
    A git session owns the index file it does operations on. We also use this
    class to abstract away the low level details of git command flags. (And in
    the future, this could be where we plug in libgit2.)'''

    def __init__(self, git_dir, index_file, working_copy):
        self.git_dir = git_dir
        self.index_file = index_file
        self.working_copy = working_copy

    @asyncio.coroutine
    def git(self, *args, input=None, output_mode=TEXT_MODE):
        global DEBUG_GIT_COMMAND_COUNT
        DEBUG_GIT_COMMAND_COUNT += 1
        command = ['git']
        command.append('--git-dir=' + self.git_dir)
        if self.working_copy:
            command.append("--work-tree=" + self.working_copy)
        command.extend(args)
        if isinstance(input, str):
            input = input.encode()
        process = yield from asyncio.subprocess.create_subprocess_exec(
            *command,
            env=self.git_env(),
            stdin=asyncio.subprocess.PIPE,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)
        stdout, stderr = yield from safe_communicate(process, input)
        stderr = stderr.decode()
        if output_mode == TEXT_MODE:
            stdout = stdout.decode()
            stdout = stdout.rstrip()
        if process.returncode != 0:
            raise GitError(command, process.returncode, stdout, stderr)
        return stdout

    def git_env(self):
        'Set the index file and prevent git from reading global configs.'
        env = dict(os.environ)
        for var in ["HOME", "XDG_CONFIG_HOME"]:
            env.pop(var, None)
        env["GIT_CONFIG_NOSYSTEM"] = "true"
        # Weirdly, GIT_INDEX_FILE is interpreted relative to the work tree. As
        # a workaround, we absoluteify the path.
        env["GIT_INDEX_FILE"] = os.path.abspath(self.index_file)
        return env

    @asyncio.coroutine
    def init_git_dir(self):
        yield from self.git('init', '--bare')

    @asyncio.coroutine
    def read_tree_into_index(self, tree):
        yield from self.git('read-tree', tree)

    @asyncio.coroutine
    def read_tree_and_stats_into_index(self, tree):
        yield from self.read_tree_into_index(tree)
        # Refresh all the stat() information in the index.
        try:
            # This throws an error on modified files. Suppress it.
            yield from self.git('update-index', '--refresh')
        except GitError as e:
            if 'needs update' not in e.stdout:
                # Reraise any errors we don't recognize.
                raise

    @asyncio.coroutine
    def make_tree_from_index(self):
        tree = yield from self.git('write-tree')
        return tree

    @asyncio.coroutine
    def read_working_copy_into_index(self, picks):
        # Use --force to avoid .gitignore rules. We shouldn't respect them.
        if picks:
            yield from self.git('add', '--force', '--', *picks)
        else:
            yield from self.git('add', '--all', '--force')

    @asyncio.coroutine
    def drop_paths_from_index(self, paths):
        if not paths:
            return
        ls_output = yield from self.git(
            'ls-files', '--full-name', '-z', *paths, output_mode=BINARY_MODE)
        yield from self.git('update-index', '--force-remove', '-z', '--stdin',
                            input=ls_output)

    @asyncio.coroutine
    def merge_tree_into_index(self, tree, prefix):
        # The --prefix argument to read-tree chokes on paths that contain dot
        # or dot-dot. Instead of './', it wants the empty string. Oblige it.
        # NOTE: This parameter must be forward-slash-separated, even on
        # Windows. os.path.normpath() is not correct here!
        prefix_path = pathlib.PurePosixPath(prefix)
        assert '..' not in prefix_path.parts
        prefix_arg = prefix_path.as_posix()
        prefix_arg = '' if prefix_arg == '.' else prefix_arg
        # The git docs say that a --prefix value must end in a slash. That
        # doesn't seem to be true in practice, but better safe than sorry. Note
        # that git treats '--prefix=/' as the root of the tree, so this doesn't
        # break that case.
        if not prefix_arg.endswith('/'):
            prefix_arg += '/'
        # Normally read-tree with --prefix wants to make sure changes don't
        # stomp on the working copy. The -i flag ignores the working copy.
        yield from self.git('read-tree', '-i', '--prefix', prefix_arg, tree)

    @asyncio.coroutine
    def working_copy_matches_index(self):
        diff_output = yield from self.git(
            'diff-files', output_mode=BINARY_MODE)
        return len(diff_output) == 0

    @asyncio.coroutine
    def get_modified_files_skipping_deletes(self):
        # We want to ignore deleted files, so we include every possible value
        # of --diff-filter except 'D'.
        diff_output = yield from self.git(
            'diff-files', '-z', '--name-only', '--diff-filter=ACMRTUXB')
        return [name for name in diff_output.split('\x00') if name]

    @asyncio.coroutine
    def get_new_files_in_tree(self, previous_tree, new_tree):
        added_files_output = yield from self.git(
            'diff-tree', '--diff-filter=A', '--name-only', '-r', '-z',
            previous_tree, new_tree)
        return added_files_output.split('\x00')

    @asyncio.coroutine
    def read_tree_updating_working_copy(self, tree, force):
        '''This method relies on the current working copy being clean with
        respect to the current index. The benefit of this over
        checkout_missing_files_from_index(), is that is clean up files that get
        deleted between the current tree and the new one. Without force, this
        raises an error rather than overwriting modified files.'''
        if force:
            yield from self.git('read-tree', '--reset', '-u', tree)
        else:
            yield from self.git('read-tree', '-m', '-u', tree)

    @asyncio.coroutine
    def checkout_files_from_index(self):
        'This recreates any deleted files.'
        yield from self.git('checkout-index', '--all')

    @asyncio.coroutine
    def get_info_for_path(self, tree, path):
        # --full-tree makes ls-tree ignore the cwd
        ls_output = yield from self.git(
            'ls-tree', '--full-tree', '-z', tree, path)
        ls_lines = ls_output.strip('\x00').split('\x00')
        # Remove empty lines.
        ls_lines = list(filter(None, ls_lines))
        if len(ls_lines) == 0:
            raise FileNotFoundError(
                'Path "{}" not found in tree {}.'.format(path, tree))
        assert len(ls_lines) == 1
        mode, type, sha1, name = ls_lines[0].split()
        return mode, type, sha1, name

    @asyncio.coroutine
    def read_bytes_from_file_hash(self, sha1):
        return (yield from self.git(
            'cat-file', '-p', sha1, output_mode=BINARY_MODE))

    @asyncio.coroutine
    def list_tree_entries(self, tree, path, recursive):
        # Lines in ls-tree are of the following form (note that the wide space
        # is a tab):
        # 100644 blob a2b67564ae3a7cb3237ee0ef1b7d26d70f2c213f    README.md
        entry_regex = r'(\w+) (\w+) (\w+)\t(.*)'
        command = ['ls-tree', '-z', tree]
        if path is not None:
            canonical_path = str(pathlib.PurePosixPath(path))
            command += [canonical_path]
        if recursive:
            # -t means tree entries are included in the listing.
            command += ['-r', '-t']
        output = yield from self.git(*command)
        if not output:
            return {}
        entries = {}
        for line in output.strip('\x00').split('\x00'):
            mode, type, hash, name = re.match(entry_regex, line).groups()
            if (recursive and path is not None and
                    len(name) < len(canonical_path) and type == TREE_TYPE):
                # In recursive mode, leave out the parents of the target dir.
                continue
            entries[name] = TreeEntry(mode, type, hash)
        return entries

    @asyncio.coroutine
    def make_tree_from_entries(self, entries):
        entry_format = '{} {} {}\t{}'
        input = '\x00'.join(entry_format.format(mode, type, hash, name)
                            for name, (mode, type, hash) in entries.items())
        tree = yield from self.git('mktree', '-z', input=input)
        return tree


@asyncio.coroutine
def Cache(root):
    'This is the async constructor for the _Cache class.'
    cache = _Cache(root)
    yield from cache._init_trees()
    return cache


class _Cache:
    def __init__(self, root):
        "Don't instantiate this class directly. Use the Cache() constructor."
        self.root = root
        self.plugins_root = os.path.join(root, "plugins")
        makedirs(self.plugins_root)
        self.tmp_path = os.path.join(root, "tmp")
        makedirs(self.tmp_path)
        self.keyval = KeyVal(os.path.join(root, 'keyval'), self.tmp_path)
        self.trees_path = os.path.join(root, "trees")
        self._empty_tree = None

    @asyncio.coroutine
    def _init_trees(self):
        if not os.path.exists(os.path.join(self.trees_path, 'HEAD')):
            os.makedirs(self.trees_path)
            with self.clean_git_session() as session:
                yield from session.init_git_dir()
            # Override any .gitattributes files that might be in the sync dir,
            # by writing 'info/attributes' in the bare repo. There are many
            # attributes that we might want to disable, but disabling 'text'
            # seems to take care of both 'text' and 'eol', which are the two
            # that I know can cause problems. We might need to add more
            # attributes here in the future. Note that other config files are
            # disabled in _git_env below.
            attributes_path = os.path.join(
                self.trees_path, 'info', 'attributes')
            with open(attributes_path, 'w') as attributes:
                # Disable the 'text' attribute for all files.
                attributes.write('* -text')

    @contextlib.contextmanager
    def clean_git_session(self, working_copy=None):
        with self.keyval.tmp_dir_context() as tmp_dir:
            # Git will initialize a nonexistent index file. Empty files cause
            # an error though.
            index_file = os.path.join(tmp_dir, "index")
            yield GitSession(self.trees_path, index_file, working_copy)

    def no_index_git_session(self):
        return GitSession(self.trees_path, os.devnull, os.devnull)

    @asyncio.coroutine
    def get_empty_tree(self):
        if not self._empty_tree:
            with self.clean_git_session() as session:
                self._empty_tree = yield from session.make_tree_from_index()
        return self._empty_tree

    @asyncio.coroutine
    def import_tree(self, src, *, picks=None, excludes=None):
        if not os.path.exists(src):
            raise RuntimeError('import tree called on nonexistent path ' + src)
        with self.clean_git_session(src) as session:
            yield from session.read_working_copy_into_index(picks)

            # We want to avoid ever importing a .peru directory. This is a
            # security/correctness issue similar to git's issue with .git dirs,
            # and just like git we need to watch out for case-insensitive
            # filesystems. See also:
            # https://github.com/blog/1938-vulnerability-announced-update-your-git-clients.
            full_excludes = dotperu_exclude_case_insensitive_git_globs()
            if excludes:
                full_excludes += excludes
            yield from session.drop_paths_from_index(full_excludes)

            tree = yield from session.make_tree_from_index()
            return tree

    @asyncio.coroutine
    def merge_trees(self, base_tree, merge_tree, merge_path='.'):
        with self.clean_git_session() as session:
            if base_tree:
                yield from session.read_tree_into_index(base_tree)
            try:
                yield from session.merge_tree_into_index(
                    merge_tree, merge_path)
            except GitError as e:
                raise MergeConflictError(e.stdout) from e
            unified_tree = yield from session.make_tree_from_index()
            return unified_tree

    @asyncio.coroutine
    def export_tree(self, tree, dest, previous_tree=None, *, force=False,
                    previous_index_file=None):
        '''This method is the core of `peru sync`. If the contents of "dest"
        match "previous_tree", then export_tree() updates them to match "tree".
        If not, it raises an error and doesn't touch any files.

        Because it's important for the no-op `peru sync` to be fast, we make an
        extra optimization for this case. The caller passes in the path to the
        index file used during the last sync, which should already reflect
        "previous_tree". That allows us to skip the read-tree and update-index
        calls, so all we have to do is a single diff-files operation to check
        for cleanliness.

        It's difficult to predict all the different states the index file might
        end up in under different error conditions, not only now but also in
        past and future git versions. For safety and simplicity, if any
        operation returns an error code, we delete the supplied index file.
        Right now this includes expected errors, like "sync would overwrite
        existing files," and unexpected errors, like "index is on fire."'''

        tree = tree or (yield from self.get_empty_tree())
        previous_tree = previous_tree or (yield from self.get_empty_tree())

        if not os.path.exists(dest):
            os.makedirs(dest)

        with contextlib.ExitStack() as stack:

            # If the caller gave us an index file, create a git session around
            # it. Otherwise, create a clean one. Note that because we delete
            # the index file whenever there are errors, we also allow the
            # caller to pass in a path to a nonexistent file. In that case we
            # have to pay the cost to recreate it.
            did_refresh = False
            if previous_index_file:
                session = GitSession(
                    self.trees_path, previous_index_file, dest)
                stack.enter_context(delete_if_error(previous_index_file))
                if not os.path.exists(previous_index_file):
                    did_refresh = True
                    yield from session.read_tree_and_stats_into_index(
                        previous_tree)
            else:
                session = stack.enter_context(self.clean_git_session(dest))
                did_refresh = True
                yield from session.read_tree_and_stats_into_index(
                    previous_tree)

            # The fast path. If the previous tree is the same as the current
            # one, and no files have changed at all, short-circuit.
            if previous_tree == tree:
                if (yield from session.working_copy_matches_index()):
                    return

            # Everything below is the slow path. Some files have changed, or
            # the tree has changed, or both. If we didn't refresh the index
            # file above, we must do so now.
            if not did_refresh:
                yield from session.read_tree_and_stats_into_index(
                    previous_tree)
            modified = yield from session.get_modified_files_skipping_deletes()
            if modified and not force:
                raise DirtyWorkingCopyError(
                    'Imported files have been modified ' +
                    '(use --force to overwrite):\n\n' +
                    _format_file_lines(modified))

            # Do all the file updates and deletions needed to produce `tree`.
            try:
                yield from session.read_tree_updating_working_copy(tree, force)
            except GitError:
                # Give a more informative error if we failed because files that
                # are new in `tree` already existed in the working copy.
                new_files = yield from session.get_new_files_in_tree(
                    previous_tree, tree)
                existing_new_files = [f for f in new_files if f and
                                      os.path.exists(os.path.join(dest, f))]
                existing_new_files.sort()
                if existing_new_files:
                    raise DirtyWorkingCopyError(
                        'Imports would overwrite preexisting files '
                        '(use --force to write anyway):\n\n' +
                        _format_file_lines(existing_new_files))
                else:
                    # We must've failed for some other reason. Let the error
                    # keep going.
                    raise

            # Recreate any missing files.
            yield from session.checkout_files_from_index()

    @asyncio.coroutine
    def read_file(self, tree, path):
        # TODO: Make this handle symlinks in the tree.
        with self.clean_git_session() as session:
            mode, type, sha1, name = yield from session.get_info_for_path(
                tree, path)
            if type == 'tree':
                raise IsADirectoryError('Path "{}" in tree {} is a directory.'
                                        .format(path, tree))
            assert type == 'blob'
            return (yield from session.read_bytes_from_file_hash(sha1))

    @asyncio.coroutine
    def ls_tree(self, tree, path=None, *, recursive=False):
        session = self.no_index_git_session()
        return (yield from session.list_tree_entries(tree, path, recursive))

    @asyncio.coroutine
    def modify_tree(self, tree, modifications):
        '''The modifications are a map of the form, {path: TreeEntry}. The tree
        can be None to indicate an empty starting tree. The entries can be
        either blobs or trees, or None to indicate a deletion. The return value
        is either the hash of the resulting tree, or None if the resulting tree
        is empty. Modifications in parent directories are done before
        modifications in subdirectories below them, so for example you can
        insert a tree at a given path and also insert more new stuff beneath
        that path, without fear of overwriting the new stuff.'''

        # Read the original contents of the base tree.
        if tree is None:
            entries = {}
        else:
            entries = yield from self.ls_tree(tree, '.')

        # Separate the modifications into two groups, those that refer to
        # entries at the base of this tree (e.g. 'foo'), and those that refer
        # to entries in subtrees (e.g. 'foo/bar').
        modifications_at_base = dict()
        modifications_in_subtrees = collections.defaultdict(dict)
        for path_str, entry in modifications.items():
            # Canonicalize paths to get rid of duplicate/trailing slashes.
            path = pathlib.PurePosixPath(path_str)

            # Check for nonsense paths.
            # TODO: Maybe stop recursive calls from repeating these checks.
            if len(path.parts) == 0:
                raise ModifyTreeError('Cannot modify an empty path.')
            elif path.parts[0] == '/':
                raise ModifyTreeError('Cannot modify an absolute path.')
            elif '..' in path.parts:
                raise ModifyTreeError('.. is not allowed in tree paths.')

            if len(path.parts) == 1:
                modifications_at_base[str(path)] = entry
            else:
                first_dir = path.parts[0]
                rest = str(pathlib.PurePosixPath(*path.parts[1:]))
                modifications_in_subtrees[first_dir][rest] = entry

        # Insert or delete entries in the base tree. Note that this happens
        # before any subtree operations.
        for name, entry in modifications_at_base.items():
            if entry is None:
                entries.pop(name, None)
            else:
                entries[name] = entry

        # Recurse to compute modified subtrees. Note how we handle deletions:
        # If 'a' is a file, inserting a new file at 'a/b' will implicitly
        # delete 'a', but trying to delete 'a/b' will be a no-op and will not
        # delete 'a'.
        empty_tree = (yield from self.get_empty_tree())
        for name, sub_modifications in modifications_in_subtrees.items():
            subtree_base = None
            if name in entries and entries[name].type == TREE_TYPE:
                subtree_base = entries[name].hash
            new_subtree = yield from self.modify_tree(
                subtree_base, sub_modifications)
            if new_subtree != empty_tree:
                entries[name] = TreeEntry(TREE_MODE, TREE_TYPE, new_subtree)
            # Delete an empty tree if it was actually a tree to begin with.
            elif name in entries and entries[name].type == TREE_TYPE:
                del entries[name]

        # Return the resulting tree, or None if empty.
        if entries:
            session = self.no_index_git_session()
            tree = yield from session.make_tree_from_entries(entries)
            return tree
        else:
            return empty_tree


@contextlib.contextmanager
def delete_if_error(path):
    '''If any exception is raised inside the context, delete the file at the
    given path, and allow the exception to continue.'''
    try:
        yield
    except:
        if os.path.exists(path):
            os.remove(path)
        raise


def _format_file_lines(files):
    '''Given a list of filenames that we're about to print, limit it to a
    reasonable number of lines.'''
    LINES_TO_SHOW = 10
    if len(files) <= LINES_TO_SHOW:
        lines = '\n'.join(files)
    else:
        lines = ('\n'.join(files[:LINES_TO_SHOW-1]) +
                 '\n...{} total'.format(len(files)))
    return lines


class GitError(Exception):
    def __init__(self, command, errorcode, stdout, stderr):
        self.command = " ".join(command)
        self.errorcode = errorcode
        self.stdout = stdout
        self.stderr = stderr
        message = textwrap.dedent('''\
            git command "{}" returned error code {}.
            stdout: {}
            stderr: {}''').format(command, errorcode, stdout, stderr)
        Exception.__init__(self, message)


class ModifyTreeError(PrintableError):
    pass


class DirtyWorkingCopyError(PrintableError):
    pass


class MergeConflictError(PrintableError):
    pass

TreeEntry = collections.namedtuple('TreeEntry', ['mode', 'type', 'hash'])

BLOB_TYPE = 'blob'
TREE_TYPE = 'tree'

NONEXECUTABLE_FILE_MODE = '100644'
EXECUTABLE_FILE_MODE = '100755'
TREE_MODE = '040000'

# All possible ways to capitalize ".peru", to exclude from imported trees.
DOTPERU_CAPITALIZATIONS = [
    '.peru',
    '.Peru',
    '.pEru',
    '.peRu',
    '.perU',
    '.PEru',
    '.PeRu',
    '.PerU',
    '.pERu',
    '.pErU',
    '.peRU',
    '.PERu',
    '.PErU',
    '.PeRU',
    '.pERU',
    '.PERU',
]


def dotperu_exclude_case_insensitive_git_globs():
    """These use the glob syntax accepted by `git ls-files` (NOT our own
    glob.py). Note that ** must match at least one path component, so we have
    to use separate globs for matches at the root and matches below."""
    globs = []
    for capitalization in DOTPERU_CAPITALIZATIONS:
        globs.append(capitalization + '/**')
        globs.append('**/' + capitalization + '/**')
    return globs
